#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# Windows highlighting code adapted from color_console.py. It is copyright
# Andre Burgaud, licensed under the MIT License, and available here:
# http://www.burgaud.com/bring-colors-to-the-windows-console-with-python/

from contextlib import contextmanager
import errno
import os
import sys
try:
    from ctypes import windll, Structure, c_short, c_ushort, byref
except ImportError:  # Not on Windows
    windll = None

from robot.errors import DataError
from robot.utils import console_encode, isatty, WINDOWS


class HighlightingStream:

    def __init__(self, stream, colors='AUTO'):
        self.stream = stream
        self._highlighter = self._get_highlighter(stream, colors)

    def _get_highlighter(self, stream, colors):
        options = {'AUTO': Highlighter if isatty(stream) else NoHighlighting,
                   'ON': Highlighter,
                   'OFF': NoHighlighting,
                   'ANSI': AnsiHighlighter}
        try:
            highlighter = options[colors.upper()]
        except KeyError:
            raise DataError("Invalid console color value '%s'. Available "
                            "'AUTO', 'ON', 'OFF' and 'ANSI'." % colors)
        return highlighter(stream)

    def write(self, text, flush=True):
        self._write(console_encode(text, stream=self.stream))
        if flush:
            self.flush()

    def _write(self, text, retry=5):
        # Workaround for Windows 10 console bug:
        # https://github.com/robotframework/robotframework/issues/2709
        try:
            with self._suppress_broken_pipe_error:
                self.stream.write(text)
        except IOError as err:
            if not (WINDOWS and err.errno == 0 and retry > 0):
                raise
            self._write(text, retry-1)

    @property
    @contextmanager
    def _suppress_broken_pipe_error(self):
        try:
            yield
        except IOError as err:
            if err.errno not in (errno.EPIPE, errno.EINVAL, errno.EBADF):
                raise

    def flush(self):
        with self._suppress_broken_pipe_error:
            self.stream.flush()

    def highlight(self, text, status=None, flush=True):
        if self._must_flush_before_and_after_highlighting():
            self.flush()
            flush = True
        with self._highlighting(status or text):
            self.write(text, flush)

    def _must_flush_before_and_after_highlighting(self):
        # Must flush on Windows before and after highlighting to make sure set
        # console colors only affect the actual highlighted text. Problems
        # only encountered with Python 3, but better to be safe than sorry.
        return WINDOWS and not isinstance(self._highlighter, NoHighlighting)

    def error(self, message, level):
        self.write('[ ', flush=False)
        self.highlight(level, flush=False)
        self.write(' ] %s\n' % message)

    @contextmanager
    def _highlighting(self, status):
        highlighter = self._highlighter
        start = {'PASS': highlighter.green,
                 'FAIL': highlighter.red,
                 'ERROR': highlighter.red,
                 'WARN': highlighter.yellow,
                 'SKIP': highlighter.yellow}[status]
        start()
        try:
            yield
        finally:
            highlighter.reset()


def Highlighter(stream):
    if os.sep == '/':
        return AnsiHighlighter(stream)
    return DosHighlighter(stream) if windll else NoHighlighting(stream)


class AnsiHighlighter:
    _ANSI_GREEN = '\033[32m'
    _ANSI_RED = '\033[31m'
    _ANSI_YELLOW = '\033[33m'
    _ANSI_RESET = '\033[0m'

    def __init__(self, stream):
        self._stream = stream

    def green(self):
        self._set_color(self._ANSI_GREEN)

    def red(self):
        self._set_color(self._ANSI_RED)

    def yellow(self):
        self._set_color(self._ANSI_YELLOW)

    def reset(self):
        self._set_color(self._ANSI_RESET)

    def _set_color(self, color):
        self._stream.write(color)


class NoHighlighting(AnsiHighlighter):

    def _set_color(self, color):
        pass


class DosHighlighter:
    _FOREGROUND_GREEN = 0x2
    _FOREGROUND_RED = 0x4
    _FOREGROUND_YELLOW = 0x6
    _FOREGROUND_GREY = 0x7
    _FOREGROUND_INTENSITY = 0x8
    _BACKGROUND_MASK = 0xF0
    _STDOUT_HANDLE = -11
    _STDERR_HANDLE = -12

    def __init__(self, stream):
        self._handle = self._get_std_handle(stream)
        self._orig_colors = self._get_colors()
        self._background = self._orig_colors & self._BACKGROUND_MASK

    def green(self):
        self._set_foreground_colors(self._FOREGROUND_GREEN)

    def red(self):
        self._set_foreground_colors(self._FOREGROUND_RED)

    def yellow(self):
        self._set_foreground_colors(self._FOREGROUND_YELLOW)

    def reset(self):
        self._set_colors(self._orig_colors)

    def _get_std_handle(self, stream):
        handle = self._STDOUT_HANDLE \
            if stream is sys.__stdout__ else self._STDERR_HANDLE
        return windll.kernel32.GetStdHandle(handle)

    def _get_colors(self):
        csbi = _CONSOLE_SCREEN_BUFFER_INFO()
        ok = windll.kernel32.GetConsoleScreenBufferInfo(self._handle, byref(csbi))
        if not ok:  # Call failed, return default console colors (gray on black)
            return self._FOREGROUND_GREY
        return csbi.wAttributes

    def _set_foreground_colors(self, colors):
        self._set_colors(colors | self._FOREGROUND_INTENSITY | self._background)

    def _set_colors(self, colors):
        windll.kernel32.SetConsoleTextAttribute(self._handle, colors)


if windll:

    class _COORD(Structure):
        _fields_ = [("X", c_short),
                    ("Y", c_short)]

    class _SMALL_RECT(Structure):
        _fields_ = [("Left", c_short),
                    ("Top", c_short),
                    ("Right", c_short),
                    ("Bottom", c_short)]

    class _CONSOLE_SCREEN_BUFFER_INFO(Structure):
        _fields_ = [("dwSize", _COORD),
                    ("dwCursorPosition", _COORD),
                    ("wAttributes", c_ushort),
                    ("srWindow", _SMALL_RECT),
                    ("dwMaximumWindowSize", _COORD)]



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import sys

from .encodingsniffer import get_console_encoding, get_system_encoding
from .misc import isatty
from .robottypes import is_string
from .unic import safe_str


CONSOLE_ENCODING = get_console_encoding()
SYSTEM_ENCODING = get_system_encoding()
PYTHONIOENCODING = os.getenv('PYTHONIOENCODING')


def console_decode(string, encoding=CONSOLE_ENCODING):
    """Decodes bytes from console encoding to Unicode.

    Uses the system console encoding by default, but that can be configured
    using the `encoding` argument. In addition to the normal encodings,
    it is possible to use case-insensitive values `CONSOLE` and `SYSTEM` to
    use the system console and system encoding, respectively.

    If `string` is already Unicode, it is returned as-is.
    """
    if is_string(string):
        return string
    encoding = {'CONSOLE': CONSOLE_ENCODING,
                'SYSTEM': SYSTEM_ENCODING}.get(encoding.upper(), encoding)
    try:
        return string.decode(encoding)
    except UnicodeError:
        return safe_str(string)


def console_encode(string, encoding=None, errors='replace', stream=sys.__stdout__,
                   force=False):
    """Encodes the given string so that it can be used in the console.

    If encoding is not given, determines it based on the given stream and system
    configuration. In addition to the normal encodings, it is possible to use
    case-insensitive values `CONSOLE` and `SYSTEM` to use the system console
    and system encoding, respectively.

    Decodes bytes back to Unicode by default, because Python 3 APIs in general
    work with strings. Use `force=True` if that is not desired.
    """
    if not is_string(string):
        string = safe_str(string)
    if encoding:
        encoding = {'CONSOLE': CONSOLE_ENCODING,
                    'SYSTEM': SYSTEM_ENCODING}.get(encoding.upper(), encoding)
    else:
        encoding = _get_console_encoding(stream)
    if encoding.upper() != 'UTF-8':
        encoded = string.encode(encoding, errors)
        return encoded if force else encoded.decode(encoding)
    return string.encode(encoding, errors) if force else string


def _get_console_encoding(stream):
    encoding = getattr(stream, 'encoding', None)
    if isatty(stream):
        return encoding or CONSOLE_ENCODING
    if PYTHONIOENCODING:
        return PYTHONIOENCODING
    return encoding or SYSTEM_ENCODING


def system_decode(string):
    return string if is_string(string) else safe_str(string)


def system_encode(string):
    return string if is_string(string) else safe_str(string)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import DataError, VariableError
from robot.utils import DotDict, is_dict_like, is_list_like, NormalizedDict, type_name

from .notfound import variable_not_found
from .resolvable import GlobalVariableValue, Resolvable
from .search import is_assign


NOT_SET = object()


class VariableStore:

    def __init__(self, variables):
        self.data = NormalizedDict(ignore='_')
        self._variables = variables

    def resolve_delayed(self, item=None):
        if item:
            return self._resolve_delayed(*item)
        for name, value in list(self.data.items()):
            try:
                self._resolve_delayed(name, value)
            except DataError:
                pass

    def _resolve_delayed(self, name, value):
        if not self._is_resolvable(value):
            return value
        try:
            self.data[name] = value.resolve(self._variables)
        except DataError as err:
            # Recursive resolving may have already removed variable.
            if name in self.data:
                self.data.pop(name)
                value.report_error(err)
            variable_not_found('${%s}' % name, self.data)
        return self.data[name]

    def _is_resolvable(self, value):
        try:
            return isinstance(value, Resolvable)
        except Exception:
            return False

    def __getitem__(self, name):
        if name not in self.data:
            variable_not_found('${%s}' % name, self.data)
        return self._resolve_delayed(name, self.data[name])

    def get(self, name, default=NOT_SET, decorated=True):
        try:
            if decorated:
                name = self._undecorate(name)
            return self[name]
        except VariableError:
            if default is NOT_SET:
                raise
            return default

    def update(self, store):
        self.data.update(store.data)

    def clear(self):
        self.data.clear()

    def add(self, name, value, overwrite=True, decorated=True):
        if decorated:
            name, value = self._undecorate_and_validate(name, value)
        if (overwrite
                or name not in self.data
                or isinstance(self.data[name], GlobalVariableValue)):
            self.data[name] = value

    def _undecorate(self, name):
        if not is_assign(name):
            raise VariableError("Invalid variable name '%s'." % name)
        return name[2:-1]

    def _undecorate_and_validate(self, name, value):
        undecorated = self._undecorate(name)
        if name[0] == '@':
            if not is_list_like(value):
                self._raise_cannot_set_type(name, value, 'list')
            value = list(value)
        if name[0] == '&':
            if not is_dict_like(value):
                self._raise_cannot_set_type(name, value, 'dictionary')
            value = DotDict(value)
        return undecorated, value

    def _raise_cannot_set_type(self, name, value, expected):
        raise VariableError("Cannot set variable '%s': Expected %s-like value, got %s."
                            % (name, expected, type_name(value)))

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    def __contains__(self, name):
        return name in self.data

    def as_dict(self, decoration=True):
        if decoration:
            variables = (self._decorate(name, self[name]) for name in self)
        else:
            variables = self.data
        return NormalizedDict(variables,  ignore='_')

    def _decorate(self, name, value):
        if is_dict_like(value):
            name = '&{%s}' % name
        elif is_list_like(value):
            name = '@{%s}' % name
        else:
            name = '${%s}' % name
        return name, value



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from collections.abc import Sequence

from ..model import TestCase


if sys.version_info >= (3, 8):
    from typing import TypedDict


    class OptionalItems(TypedDict, total=False):
        args: 'Sequence[str]'
        lineno: int


    class FixtureDict(OptionalItems):
        """Dictionary containing setup or teardown info.

        :attr:`args` and :attr:`lineno` are optional.
        """
        name: str

else:
    class FixtureDict(dict):
        pass


class TestDefaults:
    """Represents default values for test related settings set in init files.

    Parsers parsing suite files can read defaults and parsers parsing init
    files can set them. The easiest way to set defaults to a test is using
    the :meth:`set_to` method.

    This class is part of the `public parser API`__. When implementing ``parse``
    or ``parse_init`` method so that they accept two arguments, the second is
    an instance of this class. If the class is needed as a type hint, it can
    be imported via :mod:`robot.running` or :mod:`robot.api.interfaces`.

    __ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#parser-interface
    """

    def __init__(self, parent: 'TestDefaults|None' = None,
                 setup: 'FixtureDict|None' = None,
                 teardown: 'FixtureDict|None' = None,
                 tags: 'Sequence[str]' = (), timeout: 'str|None' = None):
        self.parent = parent
        self.setup = setup
        self.teardown = teardown
        self.tags = tags
        self.timeout = timeout

    @property
    def setup(self) -> 'FixtureDict|None':
        """Default setup as a ``Keyword`` object or ``None`` when not set.

        Can be set also using a dictionary.
        """
        if self._setup:
            return self._setup
        if self.parent:
            return self.parent.setup
        return None

    @setup.setter
    def setup(self, setup: 'FixtureDict|None'):
        self._setup = setup

    @property
    def teardown(self) -> 'FixtureDict|None':
        """Default teardown as a ``Keyword`` object or ``None`` when not set.

        Can be set also using a dictionary.
        """
        if self._teardown:
            return self._teardown
        if self.parent:
            return self.parent.teardown
        return None

    @teardown.setter
    def teardown(self, teardown: 'FixtureDict|None'):
        self._teardown = teardown

    @property
    def tags(self) -> 'tuple[str, ...]':
        """Default tags. Can be set also as a sequence."""
        return self._tags + self.parent.tags if self.parent else self._tags

    @tags.setter
    def tags(self, tags: 'Sequence[str]'):
        self._tags = tuple(tags)

    @property
    def timeout(self) -> 'str|None':
        """Default timeout."""
        if self._timeout:
            return self._timeout
        if self.parent:
            return self.parent.timeout
        return None

    @timeout.setter
    def timeout(self, timeout: 'str|None'):
        self._timeout = timeout

    def set_to(self, test: TestCase):
        """Sets defaults to the given test.

        Tags are always added to the test. Setup, teardown and timeout are
        set only if the test does not have them set initially.
        """
        if self.tags:
            test.tags += self.tags
        if self.setup and not test.has_setup:
            test.setup.config(**self.setup)
        if self.teardown and not test.has_teardown:
            test.teardown.config(**self.teardown)
        if self.timeout and not test.timeout:
            test.timeout = self.timeout


class FileSettings:

    def __init__(self, test_defaults: 'TestDefaults|None' = None):
        self.test_defaults = test_defaults or TestDefaults()
        self.test_setup = None
        self.test_teardown = None
        self.test_tags = ()
        self.test_timeout = None
        self.test_template = None
        self.default_tags = ()
        self.keyword_tags = ()

    @property
    def test_setup(self) -> 'FixtureDict|None':
        return self._test_setup or self.test_defaults.setup

    @test_setup.setter
    def test_setup(self, setup: 'FixtureDict|None'):
        self._test_setup = setup

    @property
    def test_teardown(self) -> 'FixtureDict|None':
        return self._test_teardown or self.test_defaults.teardown

    @test_teardown.setter
    def test_teardown(self, teardown: 'FixtureDict|None'):
        self._test_teardown = teardown

    @property
    def test_tags(self) -> 'tuple[str, ...]':
        return self._test_tags + self.test_defaults.tags

    @test_tags.setter
    def test_tags(self, tags: 'Sequence[str]'):
        self._test_tags = tuple(tags)

    @property
    def test_timeout(self) -> 'str|None':
        return self._test_timeout or self.test_defaults.timeout

    @test_timeout.setter
    def test_timeout(self, timeout: 'str|None'):
        self._test_timeout = timeout

    @property
    def test_template(self) -> 'str|None':
        return self._test_template

    @test_template.setter
    def test_template(self, template: 'str|None'):
        self._test_template = template

    @property
    def default_tags(self) -> 'tuple[str, ...]':
        return self._default_tags

    @default_tags.setter
    def default_tags(self, tags: 'Sequence[str]'):
        self._default_tags = tuple(tags)

    @property
    def keyword_tags(self) -> 'tuple[str, ...]':
        return self._keyword_tags

    @keyword_tags.setter
    def keyword_tags(self, tags: 'Sequence[str]'):
        self._keyword_tags = tuple(tags)


class InitFileSettings(FileSettings):

    @FileSettings.test_setup.setter
    def test_setup(self, setup: 'FixtureDict|None'):
        self.test_defaults.setup = setup

    @FileSettings.test_teardown.setter
    def test_teardown(self, teardown: 'FixtureDict|None'):
        self.test_defaults.teardown = teardown

    @FileSettings.test_tags.setter
    def test_tags(self, tags: 'Sequence[str]'):
        self.test_defaults.tags = tags

    @FileSettings.test_timeout.setter
    def test_timeout(self, timeout: 'str|None'):
        self.test_defaults.timeout = timeout



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.utils import get_timestamp, NullMarkupWriter, safe_str, XmlWriter
from robot.version import get_full_version
from robot.result.visitor import ResultVisitor

from .loggerhelper import IsLogged


class XmlLogger(ResultVisitor):

    def __init__(self, path, log_level='TRACE', rpa=False, generator='Robot'):
        self._log_message_is_logged = IsLogged(log_level)
        self._error_message_is_logged = IsLogged('WARN')
        self._writer = self._get_writer(path, rpa, generator)
        self._errors = []

    def _get_writer(self, path, rpa, generator):
        if not path:
            return NullMarkupWriter()
        writer = XmlWriter(path, write_empty=False, usage='output')
        writer.start('robot', {'generator': get_full_version(generator),
                               'generated': get_timestamp(),
                               'rpa': 'true' if rpa else 'false',
                               'schemaversion': '4'})
        return writer

    def close(self):
        self.start_errors()
        for msg in self._errors:
            self._write_message(msg)
        self.end_errors()
        self._writer.end('robot')
        self._writer.close()

    def set_log_level(self, level):
        return self._log_message_is_logged.set_level(level)

    def message(self, msg):
        if self._error_message_is_logged(msg.level):
            self._errors.append(msg)

    def log_message(self, msg):
        if self._log_message_is_logged(msg.level):
            self._write_message(msg)

    def _write_message(self, msg):
        attrs = {'timestamp': msg.timestamp or 'N/A', 'level': msg.level}
        if msg.html:
            attrs['html'] = 'true'
        self._writer.element('msg', msg.message, attrs)

    def start_keyword(self, kw):
        attrs = {'name': kw.kwname, 'library': kw.libname}
        if kw.type != 'KEYWORD':
            attrs['type'] = kw.type
        if kw.sourcename:
            attrs['sourcename'] = kw.sourcename
        self._writer.start('kw', attrs)
        self._write_list('var', kw.assign)
        self._write_list('arg', [safe_str(a) for a in kw.args])
        self._write_list('tag', kw.tags)
        # Must be after tags to allow adding message when using --flattenkeywords.
        self._writer.element('doc', kw.doc)

    def end_keyword(self, kw):
        if kw.timeout:
            self._writer.element('timeout', attrs={'value': str(kw.timeout)})
        self._write_status(kw)
        self._writer.end('kw')

    def start_if(self, if_):
        self._writer.start('if')
        self._writer.element('doc', if_.doc)

    def end_if(self, if_):
        self._write_status(if_)
        self._writer.end('if')

    def start_if_branch(self, branch):
        self._writer.start('branch', {'type': branch.type,
                                      'condition': branch.condition})
        self._writer.element('doc', branch.doc)

    def end_if_branch(self, branch):
        self._write_status(branch)
        self._writer.end('branch')

    def start_for(self, for_):
        self._writer.start('for', {'flavor': for_.flavor,
                                   'start': for_.start,
                                   'mode': for_.mode,
                                   'fill': for_.fill})
        for name in for_.variables:
            self._writer.element('var', name)
        for value in for_.values:
            self._writer.element('value', value)
        self._writer.element('doc', for_.doc)

    def end_for(self, for_):
        self._write_status(for_)
        self._writer.end('for')

    def start_for_iteration(self, iteration):
        self._writer.start('iter')
        for name, value in iteration.variables.items():
            self._writer.element('var', value, {'name': name})
        self._writer.element('doc', iteration.doc)

    def end_for_iteration(self, iteration):
        self._write_status(iteration)
        self._writer.end('iter')

    def start_try(self, root):
        self._writer.start('try')

    def end_try(self, root):
        self._write_status(root)
        self._writer.end('try')

    def start_try_branch(self, branch):
        if branch.type == branch.EXCEPT:
            self._writer.start('branch', attrs={
                'type': 'EXCEPT', 'variable': branch.variable,
                'pattern_type': branch.pattern_type
            })
            self._write_list('pattern', branch.patterns)
        else:
            self._writer.start('branch', attrs={'type': branch.type})

    def end_try_branch(self, branch):
        self._write_status(branch)
        self._writer.end('branch')

    def start_while(self, while_):
        self._writer.start('while', attrs={
            'condition': while_.condition,
            'limit': while_.limit,
            'on_limit': while_.on_limit,
            'on_limit_message': while_.on_limit_message
        })
        self._writer.element('doc', while_.doc)

    def end_while(self, while_):
        self._write_status(while_)
        self._writer.end('while')

    def start_while_iteration(self, iteration):
        self._writer.start('iter')
        self._writer.element('doc', iteration.doc)

    def end_while_iteration(self, iteration):
        self._write_status(iteration)
        self._writer.end('iter')

    def start_return(self, return_):
        self._writer.start('return')
        for value in return_.values:
            self._writer.element('value', value)

    def end_return(self, return_):
        self._write_status(return_)
        self._writer.end('return')

    def start_continue(self, continue_):
        self._writer.start('continue')

    def end_continue(self, continue_):
        self._write_status(continue_)
        self._writer.end('continue')

    def start_break(self, break_):
        self._writer.start('break')

    def end_break(self, break_):
        self._write_status(break_)
        self._writer.end('break')

    def start_error(self, error):
        self._writer.start('error')
        for value in error.values:
            self._writer.element('value', value)

    def end_error(self, error):
        self._write_status(error)
        self._writer.end('error')

    def start_test(self, test):
        self._writer.start('test', {'id': test.id, 'name': test.name,
                                    'line': str(test.lineno or '')})

    def end_test(self, test):
        self._writer.element('doc', test.doc)
        self._write_list('tag', test.tags)
        if test.timeout:
            self._writer.element('timeout', attrs={'value': str(test.timeout)})
        self._write_status(test)
        self._writer.end('test')

    def start_suite(self, suite):
        attrs = {'id': suite.id, 'name': suite.name}
        if suite.source:
            attrs['source'] = str(suite.source)
        self._writer.start('suite', attrs)

    def end_suite(self, suite):
        self._writer.element('doc', suite.doc)
        for name, value in suite.metadata.items():
            self._writer.element('meta', value, {'name': name})
        self._write_status(suite)
        self._writer.end('suite')

    def start_statistics(self, stats):
        self._writer.start('statistics')

    def end_statistics(self, stats):
        self._writer.end('statistics')

    def start_total_statistics(self, total_stats):
        self._writer.start('total')

    def end_total_statistics(self, total_stats):
        self._writer.end('total')

    def start_tag_statistics(self, tag_stats):
        self._writer.start('tag')

    def end_tag_statistics(self, tag_stats):
        self._writer.end('tag')

    def start_suite_statistics(self, tag_stats):
        self._writer.start('suite')

    def end_suite_statistics(self, tag_stats):
        self._writer.end('suite')

    def visit_stat(self, stat):
        self._writer.element('stat', stat.name,
                             stat.get_attributes(values_as_strings=True))

    def start_errors(self, errors=None):
        self._writer.start('errors')

    def end_errors(self, errors=None):
        self._writer.end('errors')

    def _write_list(self, tag, items):
        for item in items:
            self._writer.element(tag, item)

    def _write_status(self, item):
        attrs = {'status': item.status, 'starttime': item.starttime or 'N/A',
                 'endtime': item.endtime or 'N/A'}
        if not (item.starttime and item.endtime):
            attrs['elapsedtime'] = str(item.elapsedtime)
        self._writer.element('status', item.message, attrs)


class FlatXmlLogger(XmlLogger):

    def __init__(self, real_xml_logger):
        super().__init__(None)
        self._writer = real_xml_logger._writer

    def start_keyword(self, kw):
        pass

    def end_keyword(self, kw):
        pass

    def start_for(self, for_):
        pass

    def end_for(self, for_):
        pass

    def start_for_iteration(self, iteration):
        pass

    def end_for_iteration(self, iteration):
        pass

    def start_if(self, if_):
        pass

    def end_if(self, if_):
        pass

    def start_if_branch(self, branch):
        pass

    def end_if_branch(self, branch):
        pass

    def start_try(self, root):
        pass

    def end_try(self, root):
        pass

    def start_try_branch(self, branch):
        pass

    def end_try_branch(self, branch):
        pass

    def start_while(self, while_):
        pass

    def end_while(self, while_):
        pass

    def start_while_iteration(self, iteration):
        pass

    def end_while_iteration(self, iteration):
        pass

    def start_break(self, break_):
        pass

    def end_break(self, break_):
        pass

    def start_continue(self, continue_):
        pass

    def end_continue(self, continue_):
        pass

    def start_return(self, return_):
        pass

    def end_return(self, return_):
        pass

    def start_error(self, error):
        pass

    def end_error(self, error):
        pass



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from enum import Enum
from typing import Union, Tuple

from robot.utils import has_args, is_union, safe_str, setter, type_repr

from .argumentconverter import ArgumentConverter
from .argumentmapper import ArgumentMapper
from .argumentresolver import ArgumentResolver
from .typevalidator import TypeValidator


class ArgumentSpec:

    def __init__(self, name=None, type='Keyword', positional_only=None,
                 positional_or_named=None, var_positional=None, named_only=None,
                 var_named=None, defaults=None, types=None):
        self.name = name
        self.type = type
        self.positional_only = positional_only or []
        self.positional_or_named = positional_or_named or []
        self.var_positional = var_positional
        self.named_only = named_only or []
        self.var_named = var_named
        self.defaults = defaults or {}
        self.types = types

    @setter
    def types(self, types):
        return TypeValidator(self).validate(types)

    @property
    def positional(self):
        return self.positional_only + self.positional_or_named

    @property
    def minargs(self):
        return len([arg for arg in self.positional if arg not in self.defaults])

    @property
    def maxargs(self):
        return len(self.positional) if not self.var_positional else sys.maxsize

    @property
    def argument_names(self):
        return (self.positional_only +
                self.positional_or_named +
                ([self.var_positional] if self.var_positional else []) +
                self.named_only +
                ([self.var_named] if self.var_named else []))

    def resolve(self, arguments, variables=None, converters=None,
                resolve_named=True, resolve_variables_until=None,
                dict_to_kwargs=False, languages=None):
        resolver = ArgumentResolver(self, resolve_named,
                                    resolve_variables_until, dict_to_kwargs)
        positional, named = resolver.resolve(arguments, variables)
        return self.convert(positional, named, converters, dry_run=not variables,
                            languages=languages)

    def convert(self, positional, named, converters=None, dry_run=False, languages=None):
        if self.types or self.defaults:
            converter = ArgumentConverter(self, converters, dry_run, languages)
            positional, named = converter.convert(positional, named)
        return positional, named

    def map(self, positional, named, replace_defaults=True):
        mapper = ArgumentMapper(self)
        return mapper.map(positional, named, replace_defaults)

    def __iter__(self):
        notset = ArgInfo.NOTSET
        get_type = (self.types or {}).get
        get_default = self.defaults.get
        for arg in self.positional_only:
            yield ArgInfo(ArgInfo.POSITIONAL_ONLY, arg,
                          get_type(arg, notset), get_default(arg, notset))
        if self.positional_only:
            yield ArgInfo(ArgInfo.POSITIONAL_ONLY_MARKER)
        for arg in self.positional_or_named:
            yield ArgInfo(ArgInfo.POSITIONAL_OR_NAMED, arg,
                          get_type(arg, notset), get_default(arg, notset))
        if self.var_positional:
            yield ArgInfo(ArgInfo.VAR_POSITIONAL, self.var_positional,
                          get_type(self.var_positional, notset))
        elif self.named_only:
            yield ArgInfo(ArgInfo.NAMED_ONLY_MARKER)
        for arg in self.named_only:
            yield ArgInfo(ArgInfo.NAMED_ONLY, arg,
                          get_type(arg, notset), get_default(arg, notset))
        if self.var_named:
            yield ArgInfo(ArgInfo.VAR_NAMED, self.var_named,
                          get_type(self.var_named, notset))

    def __bool__(self):
        return any([self.positional_only, self.positional_or_named, self.var_positional,
                    self.named_only, self.var_named])

    def __str__(self):
        return ', '.join(str(arg) for arg in self)


class ArgInfo:
    """Contains argument information. Only used by Libdoc."""
    NOTSET = object()
    POSITIONAL_ONLY = 'POSITIONAL_ONLY'
    POSITIONAL_ONLY_MARKER = 'POSITIONAL_ONLY_MARKER'
    POSITIONAL_OR_NAMED = 'POSITIONAL_OR_NAMED'
    VAR_POSITIONAL = 'VAR_POSITIONAL'
    NAMED_ONLY_MARKER = 'NAMED_ONLY_MARKER'
    NAMED_ONLY = 'NAMED_ONLY'
    VAR_NAMED = 'VAR_NAMED'

    def __init__(self, kind, name='', type=NOTSET, default=NOTSET):
        self.kind = kind
        self.name = name
        self.type = TypeInfo.from_type(type)
        self.default = default

    @property
    def required(self):
        if self.kind in (self.POSITIONAL_ONLY,
                         self.POSITIONAL_OR_NAMED,
                         self.NAMED_ONLY):
            return self.default is self.NOTSET
        return False

    @property
    def types_reprs(self):
        """Deprecated. Use :attr:`type` instead."""
        if not self.type:
            return []
        if self.type.is_union:
            return [str(t) for t in self.type.nested]
        return [str(self.type)]

    @property
    def default_repr(self):
        if self.default is self.NOTSET:
            return None
        if isinstance(self.default, Enum):
            return self.default.name
        return safe_str(self.default)

    def __str__(self):
        if self.kind == self.POSITIONAL_ONLY_MARKER:
            return '/'
        if self.kind == self.NAMED_ONLY_MARKER:
            return '*'
        ret = self.name
        if self.kind == self.VAR_POSITIONAL:
            ret = '*' + ret
        elif self.kind == self.VAR_NAMED:
            ret = '**' + ret
        if self.type:
            ret = f'{ret}: {self.type}'
            default_sep = ' = '
        else:
            default_sep = '='
        if self.default is not self.NOTSET:
            ret = f'{ret}{default_sep}{self.default_repr}'
        return ret


Type = Union[type, str, tuple, type(ArgInfo.NOTSET)]


class TypeInfo:
    """Represents argument type. Only used by Libdoc.

    With unions and parametrized types, :attr:`nested` contains nested types.
    """
    NOTSET = ArgInfo.NOTSET

    def __init__(self, type: Type = NOTSET, nested: Tuple['TypeInfo'] = ()):
        self.type = type
        self.nested = nested

    @property
    def name(self) -> str:
        if isinstance(self.type, str):
            return self.type
        return type_repr(self.type, nested=False)

    @property
    def is_union(self) -> bool:
        if isinstance(self.type, str):
            return self.type == 'Union'
        return is_union(self.type, allow_tuple=True)

    @classmethod
    def from_type(cls, type: Type) -> 'TypeInfo':
        if type is cls.NOTSET:
            return cls()
        if isinstance(type, dict):
            return cls.from_dict(type)
        if isinstance(type, (tuple, list)):
            if not type:
                return cls()
            if len(type) == 1:
                return cls(type[0])
            nested = tuple(cls.from_type(t) for t in type)
            return cls('Union', nested)
        if has_args(type):
            nested = tuple(cls.from_type(t) for t in type.__args__)
            return cls(type, nested)
        return cls(type)

    @classmethod
    def from_dict(cls, data: dict) -> 'TypeInfo':
        if not data:
            return cls()
        nested = tuple(cls.from_dict(n) for n in data['nested'])
        return cls(data['name'], nested)

    def __str__(self):
        if self.is_union:
            return ' | '.join(str(n) for n in self.nested)
        if isinstance(self.type, str):
            if self.nested:
                nested = ', '.join(str(n) for n in self.nested)
                return f'{self.name}[{nested}]'
            return self.name
        return type_repr(self.type)

    def __bool__(self):
        return self.type is not self.NOTSET



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Implements report, log, output XML, and xUnit file generation.

The public API of this package is the :class:`~.ResultWriter` class. It
can write result files based on XML output files on the file system,
as well as based on the result objects returned by
the :func:`~robot.result.resultbuilder.ExecutionResult` factory method or
an executed :class:`~robot.running.model.TestSuite`.

It is highly recommended to use the public API via the :mod:`robot.api` package.

This package is considered stable.
"""

from .resultwriter import ResultWriter



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.model import BodyItem
from robot.utils import is_list_like, is_dict_like, is_string, safe_str


class ListenerArguments:

    def __init__(self, arguments):
        self._arguments = arguments
        self._version2 = None
        self._version3 = None

    def get_arguments(self, version):
        if version == 2:
            if self._version2 is None:
                self._version2 = self._get_version2_arguments(*self._arguments)
            return self._version2
        else:
            if self._version3 is None:
                self._version3 = self._get_version3_arguments(*self._arguments)
            return self._version3

    def _get_version2_arguments(self, *arguments):
        return arguments

    def _get_version3_arguments(self, *arguments):
        return arguments

    @classmethod
    def by_method_name(cls, name, arguments):
        Arguments = {'start_suite': StartSuiteArguments,
                     'end_suite': EndSuiteArguments,
                     'start_test': StartTestArguments,
                     'end_test': EndTestArguments,
                     'start_keyword': StartKeywordArguments,
                     'end_keyword': EndKeywordArguments,
                     'log_message': MessageArguments,
                     'message': MessageArguments}.get(name, ListenerArguments)
        return Arguments(arguments)


class MessageArguments(ListenerArguments):

    def _get_version2_arguments(self, msg):
        attributes = {'timestamp': msg.timestamp,
                      'message': msg.message,
                      'level': msg.level,
                      'html': 'yes' if msg.html else 'no'}
        return attributes,

    def _get_version3_arguments(self, msg):
        return msg,


class _ListenerArgumentsFromItem(ListenerArguments):
    _attribute_names = None

    def _get_version2_arguments(self, item):
        attributes = dict((name, self._get_attribute_value(item, name))
                          for name in self._attribute_names)
        attributes.update(self._get_extra_attributes(item))
        return item.name or '', attributes

    def _get_attribute_value(self, item, name):
        value = getattr(item, name)
        if value is None:
            return ''
        return self._take_copy_of_mutable_value(value)

    def _take_copy_of_mutable_value(self, value):
        if is_dict_like(value):
            return dict(value)
        if is_list_like(value):
            return list(value)
        return value

    def _get_extra_attributes(self, item):
        return {}

    def _get_version3_arguments(self, item):
        return item.data, item.result


class StartSuiteArguments(_ListenerArgumentsFromItem):
    _attribute_names = ('id', 'longname', 'doc', 'metadata', 'starttime')

    def _get_extra_attributes(self, suite):
        return {'tests': [t.name for t in suite.tests],
                'suites': [s.name for s in suite.suites],
                'totaltests': suite.test_count,
                'source': str(suite.source or '')}


class EndSuiteArguments(StartSuiteArguments):
    _attribute_names = ('id', 'longname', 'doc', 'metadata', 'starttime',
                        'endtime', 'elapsedtime', 'status', 'message')

    def _get_extra_attributes(self, suite):
        attrs = super()._get_extra_attributes(suite)
        attrs['statistics'] = suite.stat_message
        return attrs


class StartTestArguments(_ListenerArgumentsFromItem):
    _attribute_names = ('id', 'longname', 'doc', 'tags', 'lineno', 'starttime')

    def _get_extra_attributes(self, test):
        return {'source': str(test.source or ''),
                'template': test.template or '',
                'originalname': test.data.name}


class EndTestArguments(StartTestArguments):
    _attribute_names = ('id', 'longname', 'doc', 'tags', 'lineno', 'starttime',
                        'endtime', 'elapsedtime', 'status', 'message')


class StartKeywordArguments(_ListenerArgumentsFromItem):
    _attribute_names = ('doc', 'assign', 'tags', 'lineno', 'type', 'status', 'starttime')
    _type_attributes = {
        BodyItem.FOR: ('variables', 'flavor', 'values'),
        BodyItem.IF: ('condition',),
        BodyItem.ELSE_IF: ('condition',),
        BodyItem.EXCEPT: ('patterns', 'pattern_type', 'variable'),
        BodyItem.WHILE: ('condition', 'limit', 'on_limit_message'),
        BodyItem.RETURN: ('values',),
        BodyItem.ITERATION: ('variables',)
    }
    _for_flavor_attributes = {
        'IN ENUMERATE': ('start',),
        'IN ZIP': ('mode', 'fill')
    }

    def _get_extra_attributes(self, kw):
        attrs = {'kwname': kw.kwname or '',
                 'libname': kw.libname or '',
                 'args': [a if is_string(a) else safe_str(a) for a in kw.args],
                 'source': str(kw.source or '')}
        if kw.type in self._type_attributes:
            for name in self._type_attributes[kw.type]:
                if hasattr(kw, name):
                    attrs[name] = self._get_attribute_value(kw, name)
        if kw.type == BodyItem.FOR:
            for name in self._for_flavor_attributes.get(kw.flavor, ()):
                attrs[name] = self._get_attribute_value(kw, name)
        return attrs


class EndKeywordArguments(StartKeywordArguments):
    _attribute_names = ('doc', 'assign', 'tags', 'lineno', 'type', 'status',
                        'starttime', 'endtime', 'elapsedtime')



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.output.xmllogger import XmlLogger


class OutputWriter(XmlLogger):

    def __init__(self, output, rpa=False):
        XmlLogger.__init__(self, output, rpa=rpa, generator='Rebot')

    def start_message(self, msg):
        self._write_message(msg)

    def close(self):
        self._writer.end('robot')
        self._writer.close()

    def end_result(self, result):
        self.close()



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import PassExecution
from robot.model import TagPatterns
from robot.utils import html_escape, test_or_task


class Failure:

    def __init__(self):
        self.setup = None
        self.test = None
        self.teardown = None
        self.setup_skipped = None
        self.test_skipped = None
        self.teardown_skipped = None

    def __bool__(self):
        return bool(
            self.setup or self.test or self.teardown or
            self.setup_skipped or self.test_skipped or self.teardown_skipped
        )


class Exit:

    def __init__(self, failure_mode=False, error_mode=False, skip_teardown_mode=False):
        self.failure_mode = failure_mode
        self.error_mode = error_mode
        self.skip_teardown_mode = skip_teardown_mode
        self.failure = False
        self.error = False
        self.fatal = False

    def failure_occurred(self, fatal=False):
        if fatal:
            self.fatal = True
        if self.failure_mode:
            self.failure = True

    def error_occurred(self):
        if self.error_mode:
            self.error = True

    @property
    def teardown_allowed(self):
        return not (self.skip_teardown_mode and self)

    def __bool__(self):
        return bool(self.failure or self.error or self.fatal)


class _ExecutionStatus:

    def __init__(self, parent, exit=None):
        self.parent = parent
        self.exit = exit if exit is not None else parent.exit
        self.failure = Failure()
        self.skipped = False
        self._teardown_allowed = False
        self._rpa = False

    @property
    def failed(self):
        return bool(self.parent and self.parent.failed or self.failure or self.exit)

    @property
    def passed(self):
        return not self.failed

    def setup_executed(self, error=None):
        if error and not isinstance(error, PassExecution):
            msg = str(error)
            if error.skip:
                self.failure.setup_skipped = msg
                self.skipped = True
            elif self._skip_on_failure():
                self.failure.test = self._skip_on_fail_msg(f'Setup failed:\n{msg}')
                self.skipped = True
            else:
                self.failure.setup = msg
                self.exit.failure_occurred(error.exit)
        self._teardown_allowed = True

    def teardown_executed(self, error=None):
        if error and not isinstance(error, PassExecution):
            msg = str(error)
            if error.skip:
                self.failure.teardown_skipped = msg
                self.skipped = True
            elif self._skip_on_failure():
                self.failure.test = self._skip_on_fail_msg(f'Teardown failed:\n{msg}')
                self.skipped = True
            else:
                self.failure.teardown = msg
                self.exit.failure_occurred(error.exit)

    def failure_occurred(self):
        self.exit.failure_occurred()

    def error_occurred(self):
        self.exit.error_occurred()

    @property
    def teardown_allowed(self):
        return self.exit.teardown_allowed and self._teardown_allowed

    @property
    def status(self):
        if self.skipped or (self.parent and self.parent.skipped):
            return 'SKIP'
        if self.failed:
            return 'FAIL'
        return 'PASS'

    def _skip_on_failure(self):
        return False

    def _skip_on_fail_msg(self, msg):
        return msg

    @property
    def message(self):
        if self.failure or self.exit:
            return self._my_message()
        if self.parent and not self.parent.passed:
            return self._parent_message()
        return ''

    def _my_message(self):
        raise NotImplementedError

    def _parent_message(self):
        return ParentMessage(self.parent).message


class SuiteStatus(_ExecutionStatus):

    def __init__(self, parent=None, exit_on_failure=False, exit_on_error=False,
                 skip_teardown_on_exit=False):
        if parent is None:
            exit = Exit(exit_on_failure, exit_on_error, skip_teardown_on_exit)
        else:
            exit = None
        super().__init__(parent, exit)

    def _my_message(self):
        return SuiteMessage(self).message


class TestStatus(_ExecutionStatus):

    def __init__(self, parent, test, skip_on_failure=None, rpa=False):
        super().__init__(parent)
        self._test = test
        self._skip_on_failure_tags = skip_on_failure
        self._rpa = rpa

    def test_failed(self, message=None, error=None):
        if error is not None:
            message = str(error)
            skip = error.skip
            fatal = error.exit
        else:
            skip = fatal = False
        if skip:
            self.test_skipped(message)
        elif self._skip_on_failure():
            self.failure.test = self._skip_on_fail_msg(message)
            self.skipped = True
        else:
            self.failure.test = message
            self.exit.failure_occurred(fatal)

    def test_skipped(self, message):
        self.skipped = True
        self.failure.test_skipped = message

    @property
    def skip_on_failure_after_tag_changes(self):
        if not self.skipped and self.failed and self._skip_on_failure():
            self.failure.test = self._skip_on_fail_msg(self.failure.test)
            self.skipped = True
            return True
        return False

    def _skip_on_failure(self):
        return (self._test.tags.robot('skip-on-failure')
                or self._skip_on_failure_tags
                and TagPatterns(self._skip_on_failure_tags).match(self._test.tags))

    def _skip_on_fail_msg(self, msg):
        return test_or_task(
            "{Test} failed but skip-on-failure mode was active and it was marked "
            "skipped.\n\nOriginal failure:\n%s" % msg, rpa=self._rpa
        )

    def _my_message(self):
        return TestMessage(self).message


class _Message:
    setup_message = NotImplemented
    setup_skipped_message = NotImplemented
    teardown_skipped_message = NotImplemented
    teardown_message = NotImplemented
    also_teardown_message = NotImplemented

    def __init__(self, status):
        self.failure = status.failure
        self.skipped = status.skipped

    @property
    def message(self):
        message = self._get_message_before_teardown()
        return self._get_message_after_teardown(message)

    def _get_message_before_teardown(self):
        if self.failure.setup_skipped:
            return self._format_setup_or_teardown_message(
                self.setup_skipped_message, self.failure.setup_skipped)
        if self.failure.setup:
            return self._format_setup_or_teardown_message(
                self.setup_message, self.failure.setup)
        return self.failure.test_skipped or self.failure.test or ''

    def _format_setup_or_teardown_message(self, prefix, message):
        if message.startswith('*HTML*'):
            prefix = '*HTML* ' + prefix
            message = message[6:].lstrip()
        return prefix % message

    def _get_message_after_teardown(self, message):
        if not (self.failure.teardown or self.failure.teardown_skipped):
            return message
        if not message:
            if self.failure.teardown:
                prefix, msg = self.teardown_message, self.failure.teardown
            else:
                prefix, msg = self.teardown_skipped_message, self.failure.teardown_skipped
            return self._format_setup_or_teardown_message(prefix, msg)
        return self._format_message_with_teardown_message(message)

    def _format_message_with_teardown_message(self, message):
        teardown = self.failure.teardown or self.failure.teardown_skipped
        if teardown.startswith('*HTML*'):
            teardown = teardown[6:].lstrip()
            if not message.startswith('*HTML*'):
                message = '*HTML* ' + html_escape(message)
        elif message.startswith('*HTML*'):
            teardown = html_escape(teardown)
        if self.failure.teardown:
            return self.also_teardown_message % (message, teardown)
        return self.also_teardown_skip_message % (teardown, message)


class TestMessage(_Message):
    setup_message = 'Setup failed:\n%s'
    teardown_message = 'Teardown failed:\n%s'
    setup_skipped_message = '%s'
    teardown_skipped_message = '%s'
    also_teardown_message = '%s\n\nAlso teardown failed:\n%s'
    also_teardown_skip_message = 'Skipped in teardown:\n%s\n\nEarlier message:\n%s'
    exit_on_fatal_message = 'Test execution stopped due to a fatal error.'
    exit_on_failure_message = \
        'Failure occurred and exit-on-failure mode is in use.'
    exit_on_error_message = 'Error occurred and exit-on-error mode is in use.'

    def __init__(self, status):
        _Message.__init__(self, status)
        self.exit = status.exit

    @property
    def message(self):
        message = super().message
        if message:
            return message
        if self.exit.failure:
            return self.exit_on_failure_message
        if self.exit.fatal:
            return self.exit_on_fatal_message
        if self.exit.error:
            return self.exit_on_error_message
        return ''


class SuiteMessage(_Message):
    setup_message = 'Suite setup failed:\n%s'
    setup_skipped_message = 'Skipped in suite setup:\n%s'
    teardown_skipped_message = 'Skipped in suite teardown:\n%s'
    teardown_message = 'Suite teardown failed:\n%s'
    also_teardown_message = '%s\n\nAlso suite teardown failed:\n%s'
    also_teardown_skip_message = 'Skipped in suite teardown:\n%s\n\nEarlier message:\n%s'


class ParentMessage(SuiteMessage):
    setup_message = 'Parent suite setup failed:\n%s'
    setup_skipped_message = 'Skipped in parent suite setup:\n%s'
    teardown_skipped_message = 'Skipped in parent suite teardown:\n%s'
    teardown_message = 'Parent suite teardown failed:\n%s'
    also_teardown_message = '%s\n\nAlso parent suite teardown failed:\n%s'

    def __init__(self, status):
        while status.parent and status.parent.failed:
            status = status.parent
        SuiteMessage.__init__(self, status)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Package with generic, reusable and extensible model classes.

This package contains, for example, :class:`~robot.model.testsuite.TestSuite`,
:class:`~robot.model.testcase.TestCase`, :class:`~robot.model.keyword.Keyword`
and :class:`~robot.model.visitor.SuiteVisitor` base classes.
These classes are extended both by :mod:`execution <robot.running.model>`
and :mod:`result <robot.result.model>` related model objects and used also
elsewhere.

This package is considered stable.
"""

from .body import BaseBody, Body, BodyItem, BaseBranches
from .configurer import SuiteConfigurer
from .control import Break, Continue, Error, For, If, IfBranch, Return, Try, TryBranch, While
from .fixture import create_fixture
from .itemlist import ItemList
from .keyword import Keyword, Keywords
from .message import Message, Messages
from .modelobject import DataDict, ModelObject
from .modifier import ModelModifier
from .namepatterns import SuiteNamePatterns, TestNamePatterns
from .statistics import Statistics
from .tags import Tags, TagPattern, TagPatterns
from .testcase import TestCase, TestCases
from .testsuite import TestSuite, TestSuites
from .totalstatistics import TotalStatistics, TotalStatisticsBuilder
from .visitor import SuiteVisitor



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os

from .encoding import system_decode as decode, system_encode as encode


def get_env_var(name, default=None):
    try:
        value = os.environ[encode(name)]
    except KeyError:
        return default
    else:
        return decode(value)


def set_env_var(name, value):
    os.environ[encode(name)] = encode(value)


def del_env_var(name):
    value = get_env_var(name)
    if value is not None:
        del os.environ[encode(name)]
    return value


def get_env_vars(upper=os.sep != '/'):
    # by default, name is upper-cased on Windows regardless interpreter
    return dict((name if not upper else name.upper(), get_env_var(name))
                for name in (decode(name) for name in os.environ))



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings

from robot.utils import NormalizedDict


class _RunKeywordRegister:

    def __init__(self):
        self._libs = {}

    def register_run_keyword(self, libname, keyword, args_to_process,
                             deprecation_warning=True, dry_run=False):
        """Deprecated API for registering "run keyword variants".

        Registered keywords are handled specially by Robot so that:

        - Their arguments are not resolved normally (use ``args_to_process``
          to control that). This mainly means replacing variables and handling
          escapes.
        - They are not stopped by timeouts.
        - If there are conflicts with keyword names, these keywords have
          *lower* precedence than other keywords.

        This API is pretty bad and will be reimplemented in the future.
        It is thus not considered stable, but external libraries can use it
        if they really need it and are aware of forthcoming breaking changes.

        Something like this is needed at least internally also in the future.
        For external libraries we hopefully could provide a better API for
        running keywords so that they would not need this in the first place.

        For more details see the following issues and issues linked from it:
        https://github.com/robotframework/robotframework/issues/2190

        :param libname: Name of the library the keyword belongs to.
        :param keyword: Name of the keyword itself.
        :param args_to_process: How many arguments to process normally before
            passing them to the keyword. Other arguments are not touched at all.
        :param dry_run: When true, this keyword is executed in dry run. Keywords
            to actually run are got based on the ``name`` argument these
            keywords must have.
        :param deprecation_warning: Set to ``False```to avoid the warning.
        """
        if deprecation_warning:
            warnings.warn(
                "The API to register run keyword variants and to disable variable "
                "resolving in keyword arguments will change in the future. "
                "For more information see "
                "https://github.com/robotframework/robotframework/issues/2190. "
                "Use with `deprecation_warning=False` to avoid this warning.",
                UserWarning
            )
        if libname not in self._libs:
            self._libs[libname] = NormalizedDict(ignore=['_'])
        self._libs[libname][keyword] = (int(args_to_process), dry_run)

    def get_args_to_process(self, libname, kwname):
        if libname in self._libs and kwname in self._libs[libname]:
            return self._libs[libname][kwname][0]
        return -1

    def get_dry_run(self, libname, kwname):
        if libname in self._libs and kwname in self._libs[libname]:
            return self._libs[libname][kwname][1]
        return False

    def is_run_keyword(self, libname, kwname):
        return self.get_args_to_process(libname, kwname) >= 0


RUN_KW_REGISTER = _RunKeywordRegister()



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Module implementing result related model objects.

During test execution these objects are created internally by various runners.
At that time they can be inspected and modified by listeners__.

When results are parsed from XML output files after execution to be able to
create logs and reports, these objects are created by the
:func:`~.resultbuilder.ExecutionResult` factory method.
At that point they can be inspected and modified by `pre-Rebot modifiers`__.

The :func:`~.resultbuilder.ExecutionResult` factory method can also be used
by custom scripts and tools. In such usage it is often easiest to inspect and
modify these objects using the :mod:`visitor interface <robot.model.visitor>`.

If classes defined here are needed, for example, as type hints, they can
be imported via the :mod:`robot.running` module.

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#listener-interface
__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#programmatic-modification-of-results

"""

import sys
import warnings
from collections import OrderedDict
from datetime import datetime, timedelta
from itertools import chain
from pathlib import Path
from typing import Generic, Mapping, Sequence, Type, Union, TypeVar

if sys.version_info >= (3, 8):
    from typing import Literal

from robot import model
from robot.model import (BodyItem, create_fixture, DataDict, Keywords, Tags,
                         SuiteVisitor, TotalStatistics, TotalStatisticsBuilder,
                         TestCases, TestSuites)
from robot.utils import copy_signature, get_elapsed_time, KnownAtRuntime, setter

from .configurer import SuiteConfigurer
from .messagefilter import MessageFilter
from .modeldeprecation import deprecated, DeprecatedAttributesMixin
from .keywordremover import KeywordRemover
from .suiteteardownfailed import SuiteTeardownFailed, SuiteTeardownFailureHandler

IT = TypeVar('IT', bound='IfBranch|TryBranch')
FW = TypeVar('FW', bound='ForIteration|WhileIteration')

BodyItemParent = Union['TestSuite', 'TestCase', 'For', 'ForIteration', 'If', 'IfBranch',
                       'Try', 'TryBranch', 'While', 'WhileIteration', None]


class Body(model.BaseBody['Keyword', 'For', 'While', 'If', 'Try', 'Return', 'Continue',
                          'Break', 'Message', 'Error']):
    __slots__ = []


class Branches(model.BaseBranches['Keyword', 'For', 'While', 'If', 'Try', 'Return',
                                  'Continue', 'Break', 'Message', 'Error', IT]):
    __slots__ = []


class IterationType(Generic[FW]):
    """Class that wrapps `Generic` as python doesn't allow multple generic inheritance"""
    pass


class Iterations(model.BaseBody['Keyword', 'For', 'While', 'If', 'Try', 'Return',
                                'Continue', 'Break', 'Message', 'Error'], IterationType[FW]):
    __slots__ = ['iteration_class']
    iteration_type: Type[FW] = KnownAtRuntime

    def __init__(self, iteration_class: Type[FW],
                 parent: BodyItemParent = None,
                 items: 'Sequence[FW|DataDict]' = ()):
        self.iteration_class = iteration_class
        super().__init__(parent, items)

    @copy_signature(iteration_type)
    def create_iteration(self, *args, **kwargs) -> FW:
        return self._create(self.iteration_class, 'iteration_class', args, kwargs)


@Body.register
@Branches.register
@Iterations.register
class Message(model.Message):
    __slots__ = ()


class StatusMixin:
    PASS = 'PASS'
    FAIL = 'FAIL'
    SKIP = 'SKIP'
    NOT_RUN = 'NOT RUN'
    NOT_SET = 'NOT SET'
    starttime: 'str|None'
    endtime: 'str|None'
    __slots__ = ()

    @property
    def elapsedtime(self) -> int:
        """Total execution time in milliseconds.

        This attribute will be replaced by :attr:`elapsed_time` in the future.
        """
        return get_elapsed_time(self.starttime, self.endtime)

    @property
    def elapsed_time(self) -> timedelta:
        """Total execution time as a ``timedelta``.

        This attribute will replace :attr:`elapsedtime` in the future.

        New in Robot Framework 6.1.
        """
        return timedelta(milliseconds=self.elapsedtime)

    @property
    def start_time(self) -> 'datetime|None':
        """Execution start time as a ``datetime`` or as ``None`` if not set.

        This attribute will replace :attr:`starttime` in the future.

        New in Robot Framework 6.1.
        """
        return self._timestr_to_datetime(self.starttime) if self.starttime else None

    @start_time.setter
    def start_time(self, start_time: 'datetime|None'):
        self.starttime = self._datetime_to_timestr(start_time) if start_time else None

    @property
    def end_time(self) -> 'datetime|None':
        """Execution end time as a ``datetime`` or as ``None`` if not set.

        This attribute will replace :attr:`endtime` in the future.

        New in Robot Framework 6.1.
        """
        return self._timestr_to_datetime(self.endtime) if self.endtime else None

    @end_time.setter
    def end_time(self, end_time: 'datetime|None'):
        self.endtime = self._datetime_to_timestr(end_time) if end_time else None

    def _timestr_to_datetime(self, ts: str) -> datetime:
        micro = int(ts[18:]) * 1000
        return datetime(int(ts[:4]), int(ts[4:6]), int(ts[6:8]),
                        int(ts[9:11]), int(ts[12:14]), int(ts[15:17]), micro)

    def _datetime_to_timestr(self, dt: datetime) -> str:
        millis = int(round(dt.microsecond, -3) / 1000)
        return (f'{dt.year}{dt.month:02}{dt.day:02} '
                f'{dt.hour:02}:{dt.minute:02}.{dt.second:02}.{millis}')

    @property
    def passed(self) -> bool:
        """``True`` when :attr:`status` is 'PASS', ``False`` otherwise."""
        return self.status == self.PASS

    @passed.setter
    def passed(self, passed: bool):
        self.status = self.PASS if passed else self.FAIL

    @property
    def failed(self) -> bool:
        """``True`` when :attr:`status` is 'FAIL', ``False`` otherwise."""
        return self.status == self.FAIL

    @failed.setter
    def failed(self, failed: bool):
        self.status = self.FAIL if failed else self.PASS

    @property
    def skipped(self) -> bool:
        """``True`` when :attr:`status` is 'SKIP', ``False`` otherwise.

        Setting to ``False`` value is ambiguous and raises an exception.
        """
        return self.status == self.SKIP

    @skipped.setter
    def skipped(self, skipped: 'Literal[True]'):
        if not skipped:
            raise ValueError(f"`skipped` value must be truthy, got '{skipped}'.")
        self.status = self.SKIP

    @property
    def not_run(self) -> bool:
        """``True`` when :attr:`status` is 'NOT RUN', ``False`` otherwise.

        Setting to ``False`` value is ambiguous and raises an exception.
        """
        return self.status == self.NOT_RUN

    @not_run.setter
    def not_run(self, not_run: 'Literal[True]'):
        if not not_run:
            raise ValueError(f"`not_run` value must be truthy, got '{not_run}'.")
        self.status = self.NOT_RUN


class ForIteration(BodyItem, StatusMixin, DeprecatedAttributesMixin):
    """Represents one FOR loop iteration."""
    type = BodyItem.ITERATION
    body_class = Body
    repr_args = ('variables',)
    __slots__ = ['variables', 'status', 'starttime', 'endtime', 'doc']

    def __init__(self, variables: 'Mapping[str, str]|None' = None,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        self.variables = OrderedDict(variables or ())
        self.parent = parent
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc
        self.body = []

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        return self.body_class(self, body)

    def visit(self, visitor: SuiteVisitor):
        visitor.visit_for_iteration(self)

    @property
    @deprecated
    def name(self) -> str:
        return ', '.join('%s = %s' % item for item in self.variables.items())


@Body.register
class For(model.For, StatusMixin, DeprecatedAttributesMixin):
    iteration_class = ForIteration
    iterations_class = Iterations[iteration_class]
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, variables: Sequence[str] = (),
                 flavor: "Literal['IN', 'IN RANGE', 'IN ENUMERATE', 'IN ZIP']" = 'IN',
                 values: Sequence[str] = (),
                 start: 'str|None' = None,
                 mode: 'str|None' = None,
                 fill: 'str|None' = None,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(variables, flavor, values, start, mode, fill, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc

    @setter
    def body(self, iterations: 'Sequence[ForIteration|DataDict]') -> iterations_class:
        return self.iterations_class(self.iteration_class, self, iterations)

    @property
    @deprecated
    def name(self) -> str:
        variables = ' | '.join(self.variables)
        values = ' | '.join(self.values)
        for name, value in [('start', self.start),
                            ('mode', self.mode),
                            ('fill', self.fill)]:
            if value is not None:
                values += f' | {name}={value}'
        return f'{variables} {self.flavor} [ {values} ]'


class WhileIteration(BodyItem, StatusMixin, DeprecatedAttributesMixin):
    """Represents one WHILE loop iteration."""
    type = BodyItem.ITERATION
    body_class = Body
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        self.parent = parent
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        return self.body_class(self, body)

    def visit(self, visitor: SuiteVisitor):
        visitor.visit_while_iteration(self)

    @property
    @deprecated
    def name(self) -> str:
        return ''


@Body.register
class While(model.While, StatusMixin, DeprecatedAttributesMixin):
    iteration_class = WhileIteration
    iterations_class = Iterations[iteration_class]
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, condition: 'str|None' = None,
                 limit: 'str|None' = None,
                 on_limit: 'str|None' = None,
                 on_limit_message: 'str|None' = None,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(condition, limit, on_limit, on_limit_message, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc

    @setter
    def body(self, iterations: 'Sequence[WhileIteration|DataDict]') -> iterations_class:
        return self.iterations_class(self.iteration_class, self, iterations)

    @property
    @deprecated
    def name(self) -> str:
        parts = []
        if self.condition:
            parts.append(self.condition)
        if self.limit:
            parts.append(f'limit={self.limit}')
        if self.on_limit:
            parts.append(f'on_limit={self.on_limit}')
        if self.on_limit_message:
            parts.append(f'on_limit_message={self.on_limit_message}')
        return ' | '.join(parts)


class IfBranch(model.IfBranch, StatusMixin, DeprecatedAttributesMixin):
    body_class = Body
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, type: str = BodyItem.IF,
                 condition: 'str|None' = None,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(type, condition, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc

    @property
    @deprecated
    def name(self) -> str:
        return self.condition or ''


@Body.register
class If(model.If, StatusMixin, DeprecatedAttributesMixin):
    branch_class = IfBranch
    branches_class = Branches[branch_class]
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc


class TryBranch(model.TryBranch, StatusMixin, DeprecatedAttributesMixin):
    body_class = Body
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, type: str = BodyItem.TRY,
                 patterns: Sequence[str] = (),
                 pattern_type: 'str|None' = None,
                 variable: 'str|None' = None,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(type, patterns, pattern_type, variable, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc

    @property
    @deprecated
    def name(self) -> str:
        patterns = list(self.patterns)
        if self.pattern_type:
            patterns.append(f'type={self.pattern_type}')
        parts = []
        if patterns:
            parts.append(' | '.join(patterns))
        if self.variable:
            parts.append(f'AS {self.variable}')
        return ' '.join(parts)


@Body.register
class Try(model.Try, StatusMixin, DeprecatedAttributesMixin):
    branch_class = TryBranch
    branches_class = Branches[branch_class]
    __slots__ = ['status', 'starttime', 'endtime', 'doc']

    def __init__(self, status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 doc: str = '',
                 parent: BodyItemParent = None):
        super().__init__(parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.doc = doc


@Body.register
class Return(model.Return, StatusMixin, DeprecatedAttributesMixin):
    __slots__ = ['status', 'starttime', 'endtime']
    body_class = Body

    def __init__(self, values: Sequence[str] = (),
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: BodyItemParent = None):
        super().__init__(values, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Child keywords and messages as a :class:`~.Body` object.

        Typically empty. Only contains something if running RETURN has failed
        due to a syntax error or listeners have logged messages or executed
        keywords.
        """
        return self.body_class(self, body)

    @property
    @deprecated
    def args(self) -> 'tuple[str, ...]':
        return self.values

    @property
    @deprecated
    def doc(self) -> str:
        return ''


@Body.register
class Continue(model.Continue, StatusMixin, DeprecatedAttributesMixin):
    __slots__ = ['status', 'starttime', 'endtime']
    body_class = Body

    def __init__(self, status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: BodyItemParent = None):
        super().__init__(parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Child keywords and messages as a :class:`~.Body` object.

        Typically empty. Only contains something if running CONTINUE has failed
        due to a syntax error or listeners have logged messages or executed
        keywords.
        """
        return self.body_class(self, body)

    @property
    @deprecated
    def args(self) -> 'tuple[str, ...]':
        return ()

    @property
    @deprecated
    def doc(self) -> str:
        return ''


@Body.register
class Break(model.Break, StatusMixin, DeprecatedAttributesMixin):
    __slots__ = ['status', 'starttime', 'endtime']
    body_class = Body

    def __init__(self, status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: BodyItemParent = None):
        super().__init__(parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Child keywords and messages as a :class:`~.Body` object.

        Typically empty. Only contains something if running BREAK has failed
        due to a syntax error or listeners have logged messages or executed
        keywords.
        """
        return self.body_class(self, body)

    @property
    @deprecated
    def args(self) -> 'tuple[str, ...]':
        return ()

    @property
    @deprecated
    def doc(self) -> str:
        return ''


@Body.register
class Error(model.Error, StatusMixin, DeprecatedAttributesMixin):
    __slots__ = ['status', 'starttime', 'endtime']
    body_class = Body

    def __init__(self, values: Sequence[str] = (),
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: BodyItemParent = None):
        super().__init__(values, parent)
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Messages as a :class:`~.Body` object.

        Typically contains the message that caused the error.
        """
        return self.body_class(self, body)

    @property
    @deprecated
    def kwname(self) -> str:
        return self.values[0]

    @property
    @deprecated
    def args(self) -> 'tuple[str, ...]':
        return self.values[1:]

    @property
    @deprecated
    def doc(self) -> 'str':
        return ''


@Body.register
@Branches.register
@Iterations.register
class Keyword(model.Keyword, StatusMixin):
    """Represents an executed library or user keyword."""
    body_class = Body
    __slots__ = ['kwname', 'libname', 'doc', 'timeout', 'status', '_teardown',
                 'starttime', 'endtime', 'message', 'sourcename']

    def __init__(self, kwname: str = '',
                 libname: str = '',
                 doc: str = '',
                 args: Sequence[str] = (),
                 assign: Sequence[str] = (),
                 tags: Sequence[str] = (),
                 timeout: 'str|None' = None,
                 type: str = BodyItem.KEYWORD,
                 status: str = 'FAIL',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 sourcename: 'str|None' = None,
                 parent: BodyItemParent = None):
        super().__init__(None, args, assign, type, parent)
        #: Name of the keyword without library or resource name.
        self.kwname = kwname
        #: Name of the library or resource containing this keyword.
        self.libname = libname
        self.doc = doc
        self.tags = tags
        self.timeout = timeout
        self.status = status
        self.starttime = starttime
        self.endtime = endtime
        #: Keyword status message. Used only if suite teardowns fails.
        self.message = ''
        #: Original name of keyword with embedded arguments.
        self.sourcename = sourcename
        self._teardown = None
        self.body = ()

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Possible keyword body as a :class:`~.Body` object.

        Body can consist of child keywords, messages, and control structures
        such as IF/ELSE. Library keywords typically have an empty body.
        """
        return self.body_class(self, body)

    @property
    def keywords(self) -> Keywords:
        """Deprecated since Robot Framework 4.0.

        Use :attr:`body` or :attr:`teardown` instead.
        """
        keywords = self.body.filter(messages=False)
        if self.teardown:
            keywords.append(self.teardown)
        return Keywords(self, keywords)

    @keywords.setter
    def keywords(self, keywords):
        Keywords.raise_deprecation_error()

    @property
    def messages(self) -> 'list[Message]':
        """Keyword's messages.

        Starting from Robot Framework 4.0 this is a list generated from messages
        in :attr:`body`.
        """
        return self.body.filter(messages=True)    # type: ignore

    @property
    def children(self) -> 'list[BodyItem]':
        """List of child keywords and messages in creation order.

        Deprecated since Robot Framework 4.0. Use :attr:`body` instead.
        """
        warnings.warn("'Keyword.children' is deprecated. Use 'Keyword.body' instead.")
        return list(self.body)

    @property
    def name(self) -> 'str|None':
        """Keyword name in format ``libname.kwname``.

        Just ``kwname`` if :attr:`libname` is empty. In practice that is the
        case only with user keywords in the same file as the executed test case
        or test suite.

        Cannot be set directly. Set :attr:`libname` and :attr:`kwname`
        separately instead.
        """
        if not self.libname:
            return self.kwname
        return f'{self.libname}.{self.kwname}'

    @name.setter
    def name(self, name):
        if name is not None:
            raise AttributeError("Cannot set 'name' attribute directly. "
                                 "Set 'kwname' and 'libname' separately instead.")
        self.kwname = None
        self.libname = None

    @property    # Cannot use @setter because it would create teardowns recursively.
    def teardown(self) -> 'Keyword':
        """Keyword teardown as a :class:`Keyword` object.

        Teardown can be modified by setting attributes directly::

            keyword.teardown.name = 'Example'
            keyword.teardown.args = ('First', 'Second')

        Alternatively the :meth:`config` method can be used to set multiple
        attributes in one call::

            keyword.teardown.config(name='Example', args=('First', 'Second'))

        The easiest way to reset the whole teardown is setting it to ``None``.
        It will automatically recreate the underlying ``Keyword`` object::

            keyword.teardown = None

        This attribute is a ``Keyword`` object also when a keyword has no teardown
        but in that case its truth value is ``False``. If there is a need to just
        check does a keyword have a teardown, using the :attr:`has_teardown`
        attribute avoids creating the ``Keyword`` object and is thus more memory
        efficient.

        New in Robot Framework 4.0. Earlier teardown was accessed like
        ``keyword.keywords.teardown``. :attr:`has_teardown` is new in Robot
        Framework 4.1.2.
        """
        if self._teardown is None:
            self._teardown = create_fixture(self.__class__, None, self, self.TEARDOWN)
        return self._teardown

    @teardown.setter
    def teardown(self, teardown: 'Keyword|DataDict|None'):
        self._teardown = create_fixture(self.__class__, teardown, self, self.TEARDOWN)

    @property
    def has_teardown(self) -> bool:
        """Check does a keyword have a teardown without creating a teardown object.

        A difference between using ``if kw.has_teardown:`` and ``if kw.teardown:``
        is that accessing the :attr:`teardown` attribute creates a :class:`Keyword`
        object representing a teardown even when the keyword actually does not
        have one. This typically does not matter, but with bigger suite structures
        having lots of keywords it can have a considerable effect on memory usage.

        New in Robot Framework 4.1.2.
        """
        return bool(self._teardown)

    @setter
    def tags(self, tags: Sequence[str]) -> model.Tags:
        """Keyword tags as a :class:`~.model.tags.Tags` object."""
        return Tags(tags)


class TestCase(model.TestCase[Keyword], StatusMixin):
    """Represents results of a single test case.

    See the base class for documentation of attributes not documented here.
    """
    __slots__ = ['status', 'message', 'starttime', 'endtime']
    body_class = Body
    fixture_class = Keyword

    def __init__(self, name: str = '',
                 doc: str = '',
                 tags: Sequence[str] = (),
                 timeout: 'str|None' = None,
                 lineno: 'int|None' = None,
                 status: str = 'FAIL',
                 message: str = '',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: 'TestSuite|None' = None):
        super().__init__(name, doc, tags, timeout, lineno, parent)
        #: Status as a string ``PASS`` or ``FAIL``. See also :attr:`passed`.
        self.status = status
        #: Test message. Typically a failure message but can be set also when
        #: test passes.
        self.message = message
        #: Test case execution start time in format ``%Y%m%d %H:%M:%S.%f``.
        self.starttime = starttime
        #: Test case execution end time in format ``%Y%m%d %H:%M:%S.%f``.
        self.endtime = endtime

    @property
    def not_run(self) -> bool:
        return False

    @property
    def critical(self) -> bool:
        warnings.warn("'TestCase.critical' is deprecated and always returns 'True'.")
        return True

    @setter
    def body(self, body: 'Sequence[BodyItem|DataDict]') -> Body:
        """Test body as a :class:`~robot.result.Body` object."""
        return self.body_class(self, body)


class TestSuite(model.TestSuite[Keyword, TestCase], StatusMixin):
    """Represents results of a single test suite.

    See the base class for documentation of attributes not documented here.
    """
    __slots__ = ['message', 'starttime', 'endtime']
    test_class = TestCase
    fixture_class = Keyword

    def __init__(self, name: str = '',
                 doc: str = '',
                 metadata: 'Mapping[str, str]|None' = None,
                 source: 'Path|str|None' = None,
                 rpa: bool = False,
                 message: str = '',
                 starttime: 'str|None' = None,
                 endtime: 'str|None' = None,
                 parent: 'TestSuite|None' = None):
        super().__init__(name, doc, metadata, source, rpa, parent)
        #: Possible suite setup or teardown error message.
        self.message = message
        #: Suite execution start time in format ``%Y%m%d %H:%M:%S.%f``.
        self.starttime = starttime
        #: Suite execution end time in format ``%Y%m%d %H:%M:%S.%f``.
        self.endtime = endtime

    @property
    def passed(self) -> bool:
        """``True`` if no test has failed but some have passed, ``False`` otherwise."""
        return self.status == self.PASS

    @property
    def failed(self) -> bool:
        """``True`` if any test has failed, ``False`` otherwise."""
        return self.status == self.FAIL

    @property
    def skipped(self) -> bool:
        """``True`` if there are no passed or failed tests, ``False`` otherwise."""
        return self.status == self.SKIP

    @property
    def not_run(self) -> bool:
        return False

    @property
    def status(self) -> "Literal['PASS', 'SKIP', 'FAIL']":
        """'PASS', 'FAIL' or 'SKIP' depending on test statuses.

        - If any test has failed, status is 'FAIL'.
        - If no test has failed but at least some test has passed, status is 'PASS'.
        - If there are no failed or passed tests, status is 'SKIP'. This covers both
          the case when all tests have been skipped and when there are no tests.
        """
        stats = self.statistics  # Local variable avoids recreating stats.
        if stats.failed:
            return self.FAIL
        if stats.passed:
            return self.PASS
        return self.SKIP

    @property
    def statistics(self) -> TotalStatistics:
        """Suite statistics as a :class:`~robot.model.totalstatistics.TotalStatistics` object.

        Recreated every time this property is accessed, so saving the results
        to a variable and inspecting it is often a good idea::

            stats = suite.statistics
            print(stats.failed)
            print(stats.total)
            print(stats.message)
        """
        return TotalStatisticsBuilder(self, bool(self.rpa)).stats

    @property
    def full_message(self) -> str:
        """Combination of :attr:`message` and :attr:`stat_message`."""
        if not self.message:
            return self.stat_message
        return f'{self.message}\n\n{self.stat_message}'

    @property
    def stat_message(self) -> str:
        """String representation of the :attr:`statistics`."""
        return self.statistics.message

    @property
    def elapsedtime(self) -> int:
        """Total execution time in milliseconds."""
        if self.starttime and self.endtime:
            return get_elapsed_time(self.starttime, self.endtime)
        return sum(child.elapsedtime for child in
                   chain(self.suites, self.tests, (self.setup, self.teardown)))

    @setter
    def suites(self, suites: 'Sequence[TestSuite|DataDict]') -> TestSuites['TestSuite']:
        return TestSuites['TestSuite'](self.__class__, self, suites)

    def remove_keywords(self, how: str):
        """Remove keywords based on the given condition.

        :param how: What approach to use when removing keywords. Either
            ``ALL``, ``PASSED``, ``FOR``, ``WUKS``, or ``NAME:<pattern>``.

        For more information about the possible values see the documentation
        of the ``--removekeywords`` command line option.
        """
        self.visit(KeywordRemover(how))

    def filter_messages(self, log_level: str = 'TRACE'):
        """Remove log messages below the specified ``log_level``."""
        self.visit(MessageFilter(log_level))

    def configure(self, **options):
        """A shortcut to configure a suite using one method call.

        Can only be used with the root test suite.

        :param options: Passed to
            :class:`~robot.result.configurer.SuiteConfigurer` that will then
            set suite attributes, call :meth:`filter`, etc. as needed.

        Example::

            suite.configure(remove_keywords='PASSED',
                            doc='Smoke test results.')

        Not to be confused with :meth:`config` method that suites, tests,
        and keywords have to make it possible to set multiple attributes in
        one call.
        """
        super().configure()    # Parent validates is call allowed.
        self.visit(SuiteConfigurer(**options))

    def handle_suite_teardown_failures(self):
        """Internal usage only."""
        self.visit(SuiteTeardownFailureHandler())

    def suite_teardown_failed(self, message: str):
        """Internal usage only."""
        self.visit(SuiteTeardownFailed(message))

    def suite_teardown_skipped(self, message: str):
        """Internal usage only."""
        self.visit(SuiteTeardownFailed(message, skipped=True))



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

from ..lexer import Token
from ..model import (Block, Container, End, For, If, Keyword, NestedBlock,
                     Statement, TestCase, Try, While)


class Parser(ABC):
    model: Container

    def __init__(self, model: Container):
        self.model = model

    @abstractmethod
    def handles(self, statement: Statement) -> bool:
        raise NotImplementedError

    @abstractmethod
    def parse(self, statement: Statement) -> 'Parser|None':
        raise NotImplementedError


class BlockParser(Parser, ABC):
    model: Block
    unhandled_tokens = Token.HEADER_TOKENS | frozenset((Token.TESTCASE_NAME,
                                                        Token.KEYWORD_NAME))

    def __init__(self, model: Block):
        super().__init__(model)
        self.parsers: 'dict[str, type[NestedBlockParser]]' = {
            Token.FOR: ForParser,
            Token.IF: IfParser,
            Token.INLINE_IF: IfParser,
            Token.TRY: TryParser,
            Token.WHILE: WhileParser
        }

    def handles(self, statement: Statement) -> bool:
        return statement.type not in self.unhandled_tokens

    def parse(self, statement: Statement) -> 'BlockParser|None':
        parser_class = self.parsers.get(statement.type)
        if parser_class:
            model_class = parser_class.__annotations__['model']
            parser = parser_class(model_class(statement))
            self.model.body.append(parser.model)
            return parser
        self.model.body.append(statement)
        return None


class TestCaseParser(BlockParser):
    model: TestCase


class KeywordParser(BlockParser):
    model: Keyword


class NestedBlockParser(BlockParser, ABC):
    model: NestedBlock

    def __init__(self, model: NestedBlock, handle_end: bool = True):
        super().__init__(model)
        self.handle_end = handle_end

    def handles(self, statement: Statement) -> bool:
        if self.model.end:
            return False
        if statement.type == Token.END:
            return self.handle_end
        return super().handles(statement)

    def parse(self, statement: Statement) -> 'BlockParser|None':
        if isinstance(statement, End):
            self.model.end = statement
            return None
        return super().parse(statement)


class ForParser(NestedBlockParser):
    model: For


class WhileParser(NestedBlockParser):
    model: While


class IfParser(NestedBlockParser):
    model: If

    def parse(self, statement: Statement) -> 'BlockParser|None':
        if statement.type in (Token.ELSE_IF, Token.ELSE):
            parser = IfParser(If(statement), handle_end=False)
            self.model.orelse = parser.model
            return parser
        return super().parse(statement)


class TryParser(NestedBlockParser):
    model: Try

    def parse(self, statement) -> 'BlockParser|None':
        if statement.type in (Token.EXCEPT, Token.ELSE, Token.FINALLY):
            parser = TryParser(Try(statement), handle_end=False)
            self.model.next = parser.model
            return parser
        return super().parse(statement)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from os.path import basename, splitext

from robot.htmldata import HtmlFileWriter, ModelWriter, LOG, REPORT
from robot.utils import file_writer, is_string

from .jswriter import JsResultWriter, SplitLogWriter


class _LogReportWriter:
    usage = None

    def __init__(self, js_model):
        self._js_model = js_model

    def _write_file(self, path, config, template):
        outfile = file_writer(path, usage=self.usage) \
            if is_string(path) else path  # unit test hook
        with outfile:
            model_writer = RobotModelWriter(outfile, self._js_model, config)
            writer = HtmlFileWriter(outfile, model_writer)
            writer.write(template)


class LogWriter(_LogReportWriter):
    usage = 'log'

    def write(self, path, config):
        self._write_file(path, config, LOG)
        if self._js_model.split_results:
            self._write_split_logs(splitext(path)[0])

    def _write_split_logs(self, base):
        for index, (keywords, strings) in enumerate(self._js_model.split_results,
                                                    start=1):
            self._write_split_log(index, keywords, strings, '%s-%d.js' % (base, index))

    def _write_split_log(self, index, keywords, strings, path):
        with file_writer(path, usage=self.usage) as outfile:
            writer = SplitLogWriter(outfile)
            writer.write(keywords, strings, index, basename(path))


class ReportWriter(_LogReportWriter):
    usage = 'report'

    def write(self, path, config):
        self._write_file(path, config, REPORT)


class RobotModelWriter(ModelWriter):

    def __init__(self, output, model, config):
        self._output = output
        self._model = model
        self._config = config

    def write(self, line):
        JsResultWriter(self._output).write(self._model, self._config)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
import tempfile

from robot.errors import VariableError
from robot.model import Tags
from robot.output import LOGGER
from robot.utils import abspath, find_file, get_error_details, DotDict, NormalizedDict

from .resolvable import GlobalVariableValue
from .variables import Variables


class VariableScopes:

    def __init__(self, settings):
        self._global = GlobalVariables(settings)
        self._suite = None
        self._test = None
        self._scopes = [self._global]
        self._variables_set = SetVariables()

    @property
    def current(self):
        return self._scopes[-1]

    @property
    def _all_scopes(self):
        return reversed(self._scopes)

    @property
    def _scopes_until_suite(self):
        for scope in self._all_scopes:
            yield scope
            if scope is self._suite:
                break

    @property
    def _scopes_until_test(self):
        for scope in self._scopes_until_suite:
            yield scope
            if scope is self._test:
                break

    def start_suite(self):
        self._suite = self._global.copy()
        self._scopes.append(self._suite)
        self._variables_set.start_suite()
        self._variables_set.update(self._suite)

    def end_suite(self):
        self._scopes.pop()
        self._suite = self._scopes[-1] if len(self._scopes) > 1 else None
        self._variables_set.end_suite()

    def start_test(self):
        self._test = self._suite.copy()
        self._scopes.append(self._test)
        self._variables_set.start_test()

    def end_test(self):
        self._scopes.pop()
        self._test = None
        self._variables_set.end_test()

    def start_keyword(self):
        kw = self._suite.copy()
        self._variables_set.start_keyword()
        self._variables_set.update(kw)
        self._scopes.append(kw)

    def end_keyword(self):
        self._scopes.pop()
        self._variables_set.end_keyword()

    def __getitem__(self, name):
        return self.current[name]

    def __setitem__(self, name, value):
        self.current[name] = value

    def __contains__(self, name):
        return name in self.current

    def replace_list(self, items, replace_until=None, ignore_errors=False):
        return self.current.replace_list(items, replace_until, ignore_errors)

    def replace_scalar(self, items, ignore_errors=False):
        return self.current.replace_scalar(items, ignore_errors)

    def replace_string(self, string, custom_unescaper=None, ignore_errors=False):
        return self.current.replace_string(string, custom_unescaper, ignore_errors)

    def set_from_file(self, path, args, overwrite=False):
        variables = None
        for scope in self._scopes_until_suite:
            if variables is None:
                variables = scope.set_from_file(path, args, overwrite)
            else:
                scope.set_from_file(variables, overwrite=overwrite)

    def set_from_variable_table(self, variables, overwrite=False):
        for scope in self._scopes_until_suite:
            scope.set_from_variable_table(variables, overwrite)

    def resolve_delayed(self):
        for scope in self._scopes_until_suite:
            scope.resolve_delayed()

    def set_global(self, name, value):
        for scope in self._all_scopes:
            name, value = self._set_global_suite_or_test(scope, name, value)
        self._variables_set.set_global(name, value)

    def _set_global_suite_or_test(self, scope, name, value):
        scope[name] = value
        # Avoid creating new list/dict objects in different scopes.
        if name[0] != '$':
            name = '$' + name[1:]
            value = scope[name]
        return name, value

    def set_suite(self, name, value, top=False, children=False):
        if top:
            self._scopes[1][name] = value
            return
        for scope in self._scopes_until_suite:
            name, value = self._set_global_suite_or_test(scope, name, value)
        if children:
            self._variables_set.set_suite(name, value)

    def set_test(self, name, value):
        if self._test is None:
            raise VariableError('Cannot set test variable when no test is started.')
        for scope in self._scopes_until_test:
            name, value = self._set_global_suite_or_test(scope, name, value)
        self._variables_set.set_test(name, value)

    def set_keyword(self, name, value):
        self.current[name] = value
        self._variables_set.set_keyword(name, value)

    def set_local_variable(self, name, value):
        self.current[name] = value

    def as_dict(self, decoration=True):
        return self.current.as_dict(decoration=decoration)


class GlobalVariables(Variables):
    _import_by_path_ends = ('.py', '/', os.sep, '.yaml', '.yml', '.json')

    def __init__(self, settings):
        super().__init__()
        self._set_built_in_variables(settings)
        self._set_cli_variables(settings)

    def _set_cli_variables(self, settings):
        for name, args in settings.variable_files:
            try:
                if name.lower().endswith(self._import_by_path_ends):
                    name = find_file(name, file_type='Variable file')
                self.set_from_file(name, args)
            except:
                msg, details = get_error_details()
                LOGGER.error(msg)
                LOGGER.info(details)
        for varstr in settings.variables:
            try:
                name, value = varstr.split(':', 1)
            except ValueError:
                name, value = varstr, ''
            self['${%s}' % name] = value

    def _set_built_in_variables(self, settings):
        for name, value in [('${TEMPDIR}', abspath(tempfile.gettempdir())),
                            ('${EXECDIR}', abspath('.')),
                            ('${OPTIONS}', DotDict({
                                'include': Tags(settings.include),
                                'exclude': Tags(settings.exclude),
                                'skip': Tags(settings.skip),
                                'skip_on_failure': Tags(settings.skip_on_failure)
                            })),
                            ('${/}', os.sep),
                            ('${:}', os.pathsep),
                            ('${\\n}', os.linesep),
                            ('${SPACE}', ' '),
                            ('${True}', True),
                            ('${False}', False),
                            ('${None}', None),
                            ('${null}', None),
                            ('${OUTPUT_DIR}', settings.output_directory),
                            ('${OUTPUT_FILE}', settings.output or 'NONE'),
                            ('${REPORT_FILE}', settings.report or 'NONE'),
                            ('${LOG_FILE}', settings.log or 'NONE'),
                            ('${DEBUG_FILE}', settings.debug_file or 'NONE'),
                            ('${LOG_LEVEL}', settings.log_level),
                            ('${PREV_TEST_NAME}', ''),
                            ('${PREV_TEST_STATUS}', ''),
                            ('${PREV_TEST_MESSAGE}', '')]:
            self[name] = GlobalVariableValue(value)


class SetVariables:

    def __init__(self):
        self._suite = None
        self._test = None
        self._scopes = []

    def start_suite(self):
        if not self._scopes:
            self._suite = NormalizedDict(ignore='_')
        else:
            self._suite = self._scopes[-1].copy()
        self._scopes.append(self._suite)

    def end_suite(self):
        self._scopes.pop()
        self._suite = self._scopes[-1] if self._scopes else None

    def start_test(self):
        self._test = self._scopes[-1].copy()
        self._scopes.append(self._test)

    def end_test(self):
        self._test = None
        self._scopes.pop()

    def start_keyword(self):
        self._scopes.append(self._scopes[-1].copy())

    def end_keyword(self):
        self._scopes.pop()

    def set_global(self, name, value):
        for scope in self._scopes:
            if name in scope:
                scope.pop(name)

    def set_suite(self, name, value):
        self._suite[name] = value

    def set_test(self, name, value):
        for scope in reversed(self._scopes):
            scope[name] = value
            if scope is self._test:
                break

    def set_keyword(self, name, value):
        self._scopes[-1][name] = value

    def update(self, variables):
        for name, value in self._scopes[-1].items():
            variables[name] = value



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .normalizing import NormalizedDict
from .robottypes import is_string


class ConnectionCache:
    """Cache for libraries to use with concurrent connections, processes, etc.

    The cache stores the registered connections (or other objects) and allows
    switching between them using generated indices or user given aliases.
    This is useful with any library where there's need for multiple concurrent
    connections, processes, etc.

    This class is used also outside the core framework by SeleniumLibrary,
    SSHLibrary, etc. Backwards compatibility is thus important when doing changes.
    """

    def __init__(self, no_current_msg='No open connection.'):
        self._no_current = NoConnection(no_current_msg)
        self.current = self._no_current  #: Current active connection.
        self._connections = []
        self._aliases = NormalizedDict()

    @property
    def current_index(self):
        if not self:
            return None
        for index, conn in enumerate(self):
            if conn is self.current:
                return index + 1

    @current_index.setter
    def current_index(self, index):
        self.current = self._connections[index - 1] \
            if index is not None else self._no_current

    def register(self, connection, alias=None):
        """Registers given connection with optional alias and returns its index.

        Given connection is set to be the :attr:`current` connection.

        If alias is given, it must be a string. Aliases are case and space
        insensitive.

        The index of the first connection after initialization, and after
        :meth:`close_all` or :meth:`empty_cache`, is 1, second is 2, etc.
        """
        self.current = connection
        self._connections.append(connection)
        index = len(self._connections)
        if is_string(alias):
            self._aliases[alias] = index
        return index

    def switch(self, alias_or_index):
        """Switches to the connection specified by the given alias or index.

        Updates :attr:`current` and also returns its new value.

        Alias is whatever was given to :meth:`register` method and indices
        are returned by it. Index can be given either as an integer or
        as a string that can be converted to an integer. Raises an error
        if no connection with the given index or alias found.
        """
        self.current = self.get_connection(alias_or_index)
        return self.current

    def get_connection(self, alias_or_index=None):
        """Get the connection specified by the given alias or index..

        If ``alias_or_index`` is ``None``, returns the current connection
        if it is active, or raises an error if it is not.

        Alias is whatever was given to :meth:`register` method and indices
        are returned by it. Index can be given either as an integer or
        as a string that can be converted to an integer. Raises an error
        if no connection with the given index or alias found.
        """
        if alias_or_index is None:
            if not self:
                self.current.raise_error()
            return self.current
        try:
            index = self.resolve_alias_or_index(alias_or_index)
        except ValueError as err:
            raise RuntimeError(err.args[0])
        return self._connections[index-1]

    __getitem__ = get_connection

    def close_all(self, closer_method='close'):
        """Closes connections using given closer method and empties cache.

        If simply calling the closer method is not adequate for closing
        connections, clients should close connections themselves and use
        :meth:`empty_cache` afterwards.
        """
        for conn in self._connections:
            getattr(conn, closer_method)()
        self.empty_cache()
        return self.current

    def empty_cache(self):
        """Empties the connection cache.

        Indexes of the new connections starts from 1 after this.
        """
        self.current = self._no_current
        self._connections = []
        self._aliases = NormalizedDict()

    def __iter__(self):
        return iter(self._connections)

    def __len__(self):
        return len(self._connections)

    def __bool__(self):
        return self.current is not self._no_current

    def resolve_alias_or_index(self, alias_or_index):
        for resolver in self._resolve_alias, self._resolve_index, self._is_connection:
            try:
                return resolver(alias_or_index)
            except ValueError:
                pass
        raise ValueError(f"Non-existing index or alias '{alias_or_index}'.")

    def _resolve_alias(self, alias):
        if is_string(alias) and alias in self._aliases:
            return self._aliases[alias]
        raise ValueError

    def _resolve_index(self, index):
        try:
            index = int(index)
        except TypeError:
            raise ValueError
        if not 0 < index <= len(self._connections):
            raise ValueError
        return index

    def _is_connection(self, conn):
        return self._connections.index(conn) + 1


class NoConnection:

    def __init__(self, message):
        self.message = message

    def __getattr__(self, name):
        if name.startswith('__') and name.endswith('__'):
            raise AttributeError
        self.raise_error()

    def raise_error(self):
        raise RuntimeError(self.message)

    def __bool__(self):
        return False



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import os
from collections import OrderedDict
from itertools import chain

from robot.errors import DataError, KeywordError
from robot.libraries import STDLIBS
from robot.output import LOGGER, Message
from robot.utils import (RecommendationFinder, eq, find_file, is_string, normalize,
                         printable_name, seq2str2)

from .context import EXECUTION_CONTEXTS
from .importer import ImportCache, Importer
from .model import Import
from .runkwregister import RUN_KW_REGISTER
from .usererrorhandler import UserErrorHandler
from .userkeyword import UserLibrary

IMPORTER = Importer()


class Namespace:
    _default_libraries = ('BuiltIn', 'Reserved', 'Easter')
    _library_import_by_path_ends = ('.py', '/', os.sep)
    _variables_import_by_path_ends = _library_import_by_path_ends + ('.yaml', '.yml') + ('.json',)

    def __init__(self, variables, suite, resource, languages):
        LOGGER.info(f"Initializing namespace for suite '{suite.longname}'.")
        self.variables = variables
        self.languages = languages
        self._imports = resource.imports
        self._kw_store = KeywordStore(resource, languages)
        self._imported_variable_files = ImportCache()
        self._suite_name = suite.longname
        self._running_test = False

    @property
    def libraries(self):
        return self._kw_store.libraries.values()

    def handle_imports(self):
        self._import_default_libraries()
        self._handle_imports(self._imports)

    def _import_default_libraries(self):
        for name in self._default_libraries:
            self.import_library(name, notify=name == 'BuiltIn')

    def _handle_imports(self, import_settings):
        for item in import_settings:
            try:
                if not item.name:
                    raise DataError(f'{item.setting_name} setting requires value.')
                self._import(item)
            except DataError as err:
                item.report_invalid_syntax(err.message)

    def _import(self, import_setting):
        action = import_setting.select(self._import_library,
                                       self._import_resource,
                                       self._import_variables)
        action(import_setting)

    def import_resource(self, name, overwrite=True):
        self._import_resource(Import(Import.RESOURCE, name), overwrite=overwrite)

    def _import_resource(self, import_setting, overwrite=False):
        path = self._resolve_name(import_setting)
        self._validate_not_importing_init_file(path)
        if overwrite or path not in self._kw_store.resources:
            resource = IMPORTER.import_resource(path, self.languages)
            self.variables.set_from_variable_table(resource.variables, overwrite)
            user_library = UserLibrary(resource)
            self._kw_store.resources[path] = user_library
            self._handle_imports(resource.imports)
            LOGGER.imported("Resource", user_library.name,
                            importer=str(import_setting.source),
                            source=path)
        else:
            LOGGER.info(f"Resource file '{path}' already imported by "
                        f"suite '{self._suite_name}'.")

    def _validate_not_importing_init_file(self, path):
        name = os.path.splitext(os.path.basename(path))[0]
        if name.lower() == '__init__':
            raise DataError(f"Initialization file '{path}' cannot be imported as "
                            f"a resource file.")

    def import_variables(self, name, args, overwrite=False):
        self._import_variables(Import(Import.VARIABLES, name, args), overwrite)

    def _import_variables(self, import_setting, overwrite=False):
        path = self._resolve_name(import_setting)
        args = self._resolve_args(import_setting)
        if overwrite or (path, args) not in self._imported_variable_files:
            self._imported_variable_files.add((path, args))
            self.variables.set_from_file(path, args, overwrite)
            LOGGER.imported("Variables", os.path.basename(path),
                            args=list(args),
                            importer=str(import_setting.source),
                            source=path)
        else:
            msg = f"Variable file '{path}'"
            if args:
                msg += f" with arguments {seq2str2(args)}"
            LOGGER.info(f"{msg} already imported by suite '{self._suite_name}'.")

    def import_library(self, name, args=(), alias=None, notify=True):
        self._import_library(Import(Import.LIBRARY, name, args, alias), notify=notify)

    def _import_library(self, import_setting, notify=True):
        name = self._resolve_name(import_setting)
        lib = IMPORTER.import_library(name, import_setting.args,
                                      import_setting.alias, self.variables)
        if lib.name in self._kw_store.libraries:
            LOGGER.info(f"Library '{lib.name}' already imported by suite "
                        f"'{self._suite_name}'.")
            return
        if notify:
            LOGGER.imported("Library", lib.name,
                            args=list(import_setting.args),
                            originalname=lib.orig_name,
                            importer=str(import_setting.source),
                            source=lib.source)
        self._kw_store.libraries[lib.name] = lib
        lib.start_suite()
        if self._running_test:
            lib.start_test()

    def _resolve_name(self, setting):
        name = setting.name
        try:
            name = self.variables.replace_string(name)
        except DataError as err:
            self._raise_replacing_vars_failed(setting, err)
        if self._is_import_by_path(setting.type, name):
            file_type = setting.select('Library', 'Resource file', 'Variable file')
            return find_file(name, setting.directory, file_type=file_type)
        return name

    def _raise_replacing_vars_failed(self, setting, error):
        raise DataError(f"Replacing variables from setting '{setting.setting_name}' "
                        f"failed: {error}")

    def _is_import_by_path(self, import_type, path):
        if import_type == Import.LIBRARY:
            return path.lower().endswith(self._library_import_by_path_ends)
        if import_type == Import.VARIABLES:
            return path.lower().endswith(self._variables_import_by_path_ends)
        return True

    def _resolve_args(self, import_setting):
        try:
            return self.variables.replace_list(import_setting.args)
        except DataError as err:
            self._raise_replacing_vars_failed(import_setting, err)

    def set_search_order(self, new_order):
        old_order = self._kw_store.search_order
        self._kw_store.search_order = new_order
        return old_order

    def start_test(self):
        self._running_test = True
        self.variables.start_test()
        for lib in self.libraries:
            lib.start_test()

    def end_test(self):
        self.variables.end_test()
        for lib in self.libraries:
            lib.end_test()
        self._running_test = True

    def start_suite(self):
        self.variables.start_suite()

    def end_suite(self, suite):
        for lib in self.libraries:
            lib.end_suite()
        if not suite.parent:
            IMPORTER.close_global_library_listeners()
        self.variables.end_suite()

    def start_user_keyword(self):
        self.variables.start_keyword()

    def end_user_keyword(self):
        self.variables.end_keyword()

    def get_library_instance(self, libname):
        return self._kw_store.get_library(libname).get_instance()

    def get_library_instances(self):
        return dict((name, lib.get_instance())
                    for name, lib in self._kw_store.libraries.items())

    def reload_library(self, libname_or_instance):
        library = self._kw_store.get_library(libname_or_instance)
        library.reload()
        return library

    def get_runner(self, name, recommend_on_failure=True):
        try:
            return self._kw_store.get_runner(name, recommend_on_failure)
        except DataError as error:
            return UserErrorHandler(error, name)


class KeywordStore:

    def __init__(self, resource, languages):
        self.user_keywords = UserLibrary(resource, resource_file=False)
        self.libraries = OrderedDict()
        self.resources = ImportCache()
        self.search_order = ()
        self.languages = languages

    def get_library(self, name_or_instance):
        if name_or_instance is None:
            raise DataError("Library can not be None.")
        if is_string(name_or_instance):
            return self._get_lib_by_name(name_or_instance)
        return self._get_lib_by_instance(name_or_instance)

    def _get_lib_by_name(self, name):
        if name in self.libraries:
            return self.libraries[name]
        matches = [lib for lib in self.libraries.values() if eq(lib.name, name)]
        if len(matches) == 1:
            return matches[0]
        self._no_library_found(name, multiple=bool(matches))

    def _no_library_found(self, name, multiple=False):
        if multiple:
            raise DataError(f"Multiple libraries matching '{name}' found.")
        raise DataError(f"No library '{name}' found.")

    def _get_lib_by_instance(self, instance):
        for lib in self.libraries.values():
            if lib.get_instance(create=False) is instance:
                return lib
        self._no_library_found(instance)

    def get_runner(self, name, recommend=True):
        runner = self._get_runner(name)
        if runner is None:
            self._raise_no_keyword_found(name, recommend)
        return runner

    def _raise_no_keyword_found(self, name, recommend=True):
        if name.strip(': ').upper() == 'FOR':
            raise KeywordError(
                f"Support for the old FOR loop syntax has been removed. "
                f"Replace '{name}' with 'FOR', end the loop with 'END', and "
                f"remove escaping backslashes."
            )
        if name == '\\':
            raise KeywordError(
                "No keyword with name '\\' found. If it is used inside a for "
                "loop, remove escaping backslashes and end the loop with 'END'."
            )
        message = f"No keyword with name '{name}' found."
        if recommend:
            finder = KeywordRecommendationFinder(self.user_keywords,
                                                 self.libraries,
                                                 self.resources)
            raise KeywordError(finder.recommend_similar_keywords(name, message))
        else:
            raise KeywordError(message)

    def _get_runner(self, name):
        if not name:
            raise DataError('Keyword name cannot be empty.', syntax=True)
        if not is_string(name):
            raise DataError('Keyword name must be a string.', syntax=True)
        runner = self._get_runner_from_suite_file(name)
        if not runner and '.' in name:
            runner = self._get_explicit_runner(name)
        if not runner:
            runner = self._get_implicit_runner(name)
        if not runner:
            runner = self._get_bdd_style_runner(name, self.languages.bdd_prefixes)
        return runner

    def _get_bdd_style_runner(self, name, prefixes):
        parts = name.split()
        for index in range(1, len(parts)):
            prefix = ' '.join(parts[:index]).title()
            if prefix in prefixes:
                runner = self._get_runner(' '.join(parts[index:]))
                if runner:
                    runner = copy.copy(runner)
                    runner.name = name
                    return runner
        return None

    def _get_implicit_runner(self, name):
        return (self._get_runner_from_resource_files(name) or
                self._get_runner_from_libraries(name))

    def _get_runner_from_suite_file(self, name):
        if name not in self.user_keywords.handlers:
            return None
        handlers = self.user_keywords.handlers.get_handlers(name)
        if len(handlers) > 1:
            handlers = self._select_best_matches(handlers)
            if len(handlers) > 1:
                self._raise_multiple_keywords_found(handlers, name)
        runner = handlers[0].create_runner(name, self.languages)
        ctx = EXECUTION_CONTEXTS.current
        caller = ctx.user_keywords[-1] if ctx.user_keywords else ctx.test
        if caller and runner.source != caller.source:
            if self._exists_in_resource_file(name, caller.source):
                message = (
                    f"Keyword '{caller.longname}' called keyword '{name}' that exists "
                    f"both in the same resource file as the caller and in the suite "
                    f"file using that resource. The keyword in the suite file is used "
                    f"now, but this will change in Robot Framework 7.0."
                )
                runner.pre_run_messages += Message(message, level='WARN'),
        return runner

    def _select_best_matches(self, handlers):
        # "Normal" matches are considered exact and win over embedded matches.
        normal = [hand for hand in handlers if not hand.supports_embedded_args]
        if normal:
            return normal
        matches = [hand for hand in handlers
                   if not self._is_worse_match_than_others(hand, handlers)]
        return matches or handlers

    def _is_worse_match_than_others(self, candidate, alternatives):
        for other in alternatives:
            if (candidate is not other
                    and self._is_better_match(other, candidate)
                    and not self._is_better_match(candidate, other)):
                return True
        return False

    def _is_better_match(self, candidate, other):
        # Embedded match is considered better than another if the other matches
        # it, but it doesn't match the other.
        return other.matches(candidate.name) and not candidate.matches(other.name)

    def _exists_in_resource_file(self, name, source):
        for resource in self.resources.values():
            if resource.source == source and name in resource.handlers:
                return True
        return False

    def _get_runner_from_resource_files(self, name):
        handlers = [handler for res in self.resources.values()
                    for handler in res.handlers_for(name)]
        if not handlers:
            return None
        if len(handlers) > 1:
            handlers = self._filter_based_on_search_order(handlers)
            if len(handlers) > 1:
                handlers = self._prioritize_same_file_or_public(handlers)
                if len(handlers) > 1:
                    handlers = self._select_best_matches(handlers)
                    if len(handlers) > 1:
                        self._raise_multiple_keywords_found(handlers, name)
        return handlers[0].create_runner(name, self.languages)

    def _get_runner_from_libraries(self, name):
        handlers = [handler for lib in self.libraries.values()
                    for handler in lib.handlers_for(name)]
        if not handlers:
            return None
        pre_run_message = None
        if len(handlers) > 1:
            handlers = self._filter_based_on_search_order(handlers)
            if len(handlers) > 1:
                handlers = self._select_best_matches(handlers)
                if len(handlers) > 1:
                    handlers, pre_run_message = self._filter_stdlib_handler(handlers)
                    if len(handlers) > 1:
                        self._raise_multiple_keywords_found(handlers, name)
        runner = handlers[0].create_runner(name, self.languages)
        if pre_run_message:
            runner.pre_run_messages += (pre_run_message,)
        return runner

    def _prioritize_same_file_or_public(self, handlers):
        user_keywords = EXECUTION_CONTEXTS.current.user_keywords
        if user_keywords:
            parent_source = user_keywords[-1].source
            matches = [h for h in handlers if h.source == parent_source]
            if matches:
                return matches
        matches = [handler for handler in handlers if not handler.private]
        return matches or handlers

    def _filter_based_on_search_order(self, handlers):
        for libname in self.search_order:
            matches = [hand for hand in handlers if eq(libname, hand.libname)]
            if matches:
                return matches
        return handlers

    def _filter_stdlib_handler(self, handlers):
        warning = None
        if len(handlers) != 2:
            return handlers, warning
        stdlibs_without_remote = STDLIBS - {'Remote'}
        if handlers[0].library.orig_name in stdlibs_without_remote:
            standard, custom = handlers
        elif handlers[1].library.orig_name in stdlibs_without_remote:
            custom, standard = handlers
        else:
            return handlers, warning
        if not RUN_KW_REGISTER.is_run_keyword(custom.library.orig_name, custom.name):
            warning = self._custom_and_standard_keyword_conflict_warning(custom, standard)
        return [custom], warning

    def _custom_and_standard_keyword_conflict_warning(self, custom, standard):
        custom_with_name = standard_with_name = ''
        if custom.library.name != custom.library.orig_name:
            custom_with_name = f" imported as '{custom.library.name}'"
        if standard.library.name != standard.library.orig_name:
            standard_with_name = f" imported as '{standard.library.name}'"
        return Message(
            f"Keyword '{standard.name}' found both from a custom library "
            f"'{custom.library.orig_name}'{custom_with_name} and a standard library "
            f"'{standard.library.orig_name}'{standard_with_name}. The custom keyword "
            f"is used. To select explicitly, and to get rid of this warning, use "
            f"either '{custom.longname}' or '{standard.longname}'.", level='WARN'
        )

    def _get_explicit_runner(self, name):
        handlers_and_names = []
        for owner_name, kw_name in self._get_owner_and_kw_names(name):
            for owner in chain(self.libraries.values(), self.resources.values()):
                if eq(owner.name, owner_name) and kw_name in owner.handlers:
                    for handler in owner.handlers.get_handlers(kw_name):
                        handlers_and_names.append((handler, kw_name))
        if not handlers_and_names:
            return None
        if len(handlers_and_names) == 1:
            handler, kw_name = handlers_and_names[0]
        else:
            handlers = [h for h, n in handlers_and_names]
            matches = self._select_best_matches(handlers)
            if len(matches) > 1:
                self._raise_multiple_keywords_found(handlers, name, implicit=False)
            handler, kw_name = handlers_and_names[handlers.index(matches[0])]
        return handler.create_runner(kw_name, self.languages)

    def _get_owner_and_kw_names(self, full_name):
        tokens = full_name.split('.')
        return [('.'.join(tokens[:index]), '.'.join(tokens[index:]))
                for index in range(1, len(tokens))]

    def _raise_multiple_keywords_found(self, handlers, name, implicit=True):
        if any(hand.supports_embedded_args for hand in handlers):
            error = f"Multiple keywords matching name '{name}' found"
        else:
            error = f"Multiple keywords with name '{name}' found"
            if implicit:
                error += ". Give the full name of the keyword you want to use"
        names = sorted(hand.longname for hand in handlers)
        raise KeywordError('\n    '.join([error+':'] + names))


class KeywordRecommendationFinder:

    def __init__(self, user_keywords, libraries, resources):
        self.user_keywords = user_keywords
        self.libraries = libraries
        self.resources = resources

    def recommend_similar_keywords(self, name, message):
        """Return keyword names similar to `name`."""
        candidates = self._get_candidates('.' in name)
        finder = RecommendationFinder(
            lambda name: normalize(candidates.get(name, name), ignore='_')
        )
        return finder.find_and_format(name, candidates, message,
                                      check_missing_argument_separator=True)

    @staticmethod
    def format_recommendations(message, recommendations):
        return RecommendationFinder().format(message, recommendations)

    def _get_candidates(self, use_full_name):
        names = {}
        for owner, name in self._get_all_handler_names():
            full_name = f'{owner}.{name}' if owner else name
            names[full_name] = full_name if use_full_name else name
        return names

    def _get_all_handler_names(self):
        """Return a list of `(library_name, handler_name)` tuples."""
        handlers = [('', printable_name(handler.name, True))
                    for handler in self.user_keywords.handlers]
        for library in chain(self.libraries.values(), self.resources.values()):
            if library.name != 'Reserved':
                handlers.extend(
                    ((library.name or '',
                      printable_name(handler.name, code_style=True))
                     for handler in library.handlers))
        return sorted(handlers)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import copy
import os.path
import os

from robot.output import LOGGER
from robot.errors import FrameworkError, DataError
from robot.utils import normpath, seq2str, seq2str2, is_string

from .builder import ResourceFileBuilder
from .handlerstore import HandlerStore
from .testlibraries import TestLibrary


RESOURCE_EXTENSIONS = {'.resource', '.robot', '.txt', '.tsv', '.rst', '.rest',
                       '.json', '.rsrc'}


class Importer:

    def __init__(self):
        self._library_cache = ImportCache()
        self._resource_cache = ImportCache()

    def reset(self):
        self.__init__()

    def close_global_library_listeners(self):
        for lib in self._library_cache.values():
            lib.close_global_listeners()

    def import_library(self, name, args, alias, variables):
        lib = TestLibrary(name, args, variables, create_handlers=False)
        positional, named = lib.positional_args, lib.named_args
        lib = self._import_library(name, positional, named, lib)
        if alias:
            alias = variables.replace_scalar(alias)
            lib = self._copy_library(lib, alias)
            LOGGER.info(f"Imported library '{name}' with name '{alias}'.")
        return lib

    def import_resource(self, path, lang=None):
        self._validate_resource_extension(path)
        if path in self._resource_cache:
            LOGGER.info(f"Found resource file '{path}' from cache.")
        else:
            resource = ResourceFileBuilder(lang=lang).build(path)
            self._resource_cache[path] = resource
        return self._resource_cache[path]

    def _validate_resource_extension(self, path):
        extension = os.path.splitext(path)[1]
        if extension.lower() not in RESOURCE_EXTENSIONS:
            extensions = seq2str(sorted(RESOURCE_EXTENSIONS))
            raise DataError(f"Invalid resource file extension '{extension}'. "
                            f"Supported extensions are {extensions}.")

    def _import_library(self, name, positional, named, lib):
        args = positional + [f'{name}={value}' for name, value in named]
        key = (name, positional, named)
        if key in self._library_cache:
            LOGGER.info(f"Found library '{name}' with arguments {seq2str2(args)} "
                        f"from cache.")
            return self._library_cache[key]
        lib.create_handlers()
        self._library_cache[key] = lib
        self._log_imported_library(name, args, lib)
        return lib

    def _log_imported_library(self, name, args, lib):
        type = lib.__class__.__name__.replace('Library', '').lower()[1:]
        listener = ', with listener' if lib.has_listener else ''
        LOGGER.info(f"Imported library '{name}' with arguments {seq2str2(args)} "
                    f"(version {lib.version or '<unknown>'}, {type} type, "
                    f"{lib.scope} scope, {len(lib)} keywords{listener}).")
        if not lib:
            LOGGER.warn(f"Imported library '{name}' contains no keywords.")

    def _copy_library(self, orig, name):
        # This is pretty ugly. Hopefully we can remove cache and copying
        # altogether in 3.0 and always just re-import libraries:
        # https://github.com/robotframework/robotframework/issues/2106
        lib = copy.copy(orig)
        lib.name = name
        lib.scope = type(lib.scope)(lib)
        lib.reset_instance()
        lib.handlers = HandlerStore()
        for handler in orig.handlers._normal.values():
            handler = copy.copy(handler)
            handler.library = lib
            lib.handlers.add(handler)
        for handler in orig.handlers._embedded:
            handler = copy.copy(handler)
            handler.library = lib
            lib.handlers.add(handler, embedded=True)
        return lib


class ImportCache:
    """Keeps track on and optionally caches imported items.

    Handles paths in keys case-insensitively on case-insensitive OSes.
    Unlike dicts, this storage accepts mutable values in keys.
    """

    def __init__(self):
        self._keys = []
        self._items = []

    def __setitem__(self, key, item):
        if not is_string(key) and not isinstance(key, tuple):
            raise FrameworkError('Invalid key for ImportCache')
        key = self._norm_path_key(key)
        if key not in self._keys:
            self._keys.append(key)
            self._items.append(item)
        else:
            self._items[self._keys.index(key)] = item

    def add(self, key, item=None):
        self.__setitem__(key, item)

    def __getitem__(self, key):
        key = self._norm_path_key(key)
        if key not in self._keys:
            raise KeyError
        return self._items[self._keys.index(key)]

    def __contains__(self, key):
        return self._norm_path_key(key) in self._keys

    def values(self):
        return self._items

    def _norm_path_key(self, key):
        if self._is_path(key):
            return normpath(key, case_normalize=True)
        if isinstance(key, tuple):
            return tuple(self._norm_path_key(k) for k in key)
        return key

    def _is_path(self, key):
        return is_string(key) and os.path.isabs(key) and os.path.exists(key)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import date, datetime, timedelta
from decimal import Decimal
from pathlib import Path
from typing import Any


STANDARD_TYPE_DOCS = {
    Any: '''\
Any value is accepted. No conversion is done.
''',
    bool: '''\
Strings ``TRUE``, ``YES``, ``ON`` and ``1`` are converted to Boolean ``True``,
the empty string as well as strings ``FALSE``, ``NO``, ``OFF`` and ``0``
are converted to Boolean ``False``, and the string ``NONE`` is converted
to the Python ``None`` object. Other strings and other accepted values are
passed as-is, allowing keywords to handle them specially if
needed. All string comparisons are case-insensitive.

Examples: ``TRUE`` (converted to ``True``), ``off`` (converted to ``False``),
``example`` (used as-is)
''',
    int: '''\
Conversion is done using Python's [https://docs.python.org/library/functions.html#int|int]
built-in function. Floating point
numbers are accepted only if they can be represented as integers exactly.
For example, ``1.0`` is accepted and ``1.1`` is not.

Starting from RF 4.1, it is possible to use hexadecimal, octal and binary
numbers by prefixing values with ``0x``, ``0o`` and ``0b``, respectively.

Starting from RF 4.1, spaces and underscores can be used as visual separators
for digit grouping purposes.

Examples: ``42``, ``-1``, ``0b1010``, ``10 000 000``, ``0xBAD_C0FFEE``
''',
    float: '''\
Conversion is done using Python's
[https://docs.python.org/library/functions.html#float|float] built-in function.

Starting from RF 4.1, spaces and underscores can be used as visual separators
for digit grouping purposes.

Examples: ``3.14``, ``2.9979e8``, ``10 000.000 01``
''',
    Decimal: '''\
Conversion is done using Python's
[https://docs.python.org/library/decimal.html#decimal.Decimal|Decimal] class.

Starting from RF 4.1, spaces and underscores can be used as visual separators
for digit grouping purposes.

Examples: ``3.14``, ``10 000.000 01``
''',
    str: 'All arguments are converted to Unicode strings.',
    bytes: '''\
Strings are converted to bytes so that each Unicode code point
below 256 is directly mapped to a matching byte. Higher code
points are not allowed. Robot Framework's ``\\xHH`` escape syntax is
convenient with bytes having non-printable values.

Examples: ``good``, ``hyvä`` (same as ``hyv\\xE4``), ``\\x00`` (the null byte)
''',
    bytearray: 'Set below to same value as `bytes`.',
    datetime: '''\
Strings are expected to be a timestamp in
[https://en.wikipedia.org/wiki/ISO_8601|ISO 8601] like
format ``YYYY-MM-DD hh:mm:ss.mmmmmm``, where any non-digit
character can be used as a separator or separators can be
omitted altogether. Additionally, only the date part is
mandatory, all possibly missing time components are considered
to be zeros.

Integers and floats are considered to represent seconds since
the [https://en.wikipedia.org/wiki/Unix_time|Unix epoch].

Examples: ``2022-02-09T16:39:43.632269``, ``2022-02-09 16:39``,
``${1644417583.632269}`` (Epoch time)
''',
    date: '''\
Strings are expected to be a timestamp in
[https://en.wikipedia.org/wiki/ISO_8601|ISO 8601] like date format
``YYYY-MM-DD``, where any non-digit character can be used as a separator
or separators can be omitted altogether. Possible time components are
only allowed if they are zeros.

Examples: ``2022-02-09``, ``2022-02-09 00:00``
''',
    timedelta: '''\
Strings are expected to represent a time interval in one of
the time formats Robot Framework supports:
- a number representing seconds like ``42`` or ``10.5``
- a time string like ``1 hour 2 seconds`` or ``1h 2s``
- a "timer" string like ``01:02`` (1 minute 2 seconds) or ``01:00:03`` (1 hour 3 seconds)

Integers and floats are considered to be seconds.

See the [https://robotframework.org/robotframework/|Robot Framework User Guide]
for more details about the supported time formats.
''',
    Path: '''\
Strings are converted [https://docs.python.org/library/pathlib.html|Path] objects.
On Windows ``/`` is converted to ``\\`` automatically.

Examples: ``/tmp/absolute/path``, ``relative/path/to/file.ext``, ``name.txt``
''',
    type(None): '''\
String ``NONE`` (case-insensitive) is converted to Python ``None`` object.
Other values cause an error.
''',
    list: '''\
Strings must be Python [https://docs.python.org/library/stdtypes.html#list|list]
literals. They are converted to actual lists using the
[https://docs.python.org/library/ast.html#ast.literal_eval|ast.literal_eval]
function. They can contain any values ``ast.literal_eval`` supports, including
lists and other containers.

If the type has nested types like ``list[int]``, items are converted
to those types automatically. This in new in Robot Framework 6.0.

Examples: ``['one', 'two']``, ``[('one', 1), ('two', 2)]``
''',
    tuple: '''\
Strings must be Python [https://docs.python.org/library/stdtypes.html#tuple|tuple]
literals. They are converted to actual tuples using the
[https://docs.python.org/library/ast.html#ast.literal_eval|ast.literal_eval]
function. They can contain any values ``ast.literal_eval`` supports, including
tuples and other containers.

If the type has nested types like ``tuple[str, int, int]``, items are converted
to those types automatically. This in new in Robot Framework 6.0.

Examples: ``('one', 'two')``, ``(('one', 1), ('two', 2))``
''',
    dict: '''\
Strings must be Python [https://docs.python.org/library/stdtypes.html#dict|dictionary]
literals. They are converted to actual dictionaries using the
[https://docs.python.org/library/ast.html#ast.literal_eval|ast.literal_eval]
function. They can contain any values ``ast.literal_eval`` supports, including
dictionaries and other containers.

If the type has nested types like ``dict[str, int]``, items are converted
to those types automatically. This in new in Robot Framework 6.0.

Examples: ``{'a': 1, 'b': 2}``, ``{'key': 1, 'nested': {'key': 2}}``
''',
    set: '''\
Strings must be Python [https://docs.python.org/library/stdtypes.html#set|set]
literals. They are converted to actual sets using the
[https://docs.python.org/library/ast.html#ast.literal_eval|ast.literal_eval]
function. They can contain any values ``ast.literal_eval`` supports.

If the type has nested types like ``set[int]``, items are converted
to those types automatically. This in new in Robot Framework 6.0.

Examples: ``{1, 2, 3, 42}``, ``set()`` (an empty set)
''',
    frozenset:  '''\
Strings must be Python [https://docs.python.org/library/stdtypes.html#set|set]
literals. They are converted to actual sets using the
[https://docs.python.org/library/ast.html#ast.literal_eval|ast.literal_eval]
function and then converted to ``frozenset`` objects. They can contain
any values ``ast.literal_eval`` supports.

If the type has nested types like ``frozenset[int]``, items are converted
to those types automatically. This in new in Robot Framework 6.0.

Examples: ``{1, 2, 3, 42}``, ``set()`` (an empty set)
'''
}

STANDARD_TYPE_DOCS[bytearray] = STANDARD_TYPE_DOCS[bytes]
